//----------------------------------------------------------------------------------------
// mesh_3d_sw.H: Contains declarations for 3D node, face, and prism shallow-water objects
//----------------------------------------------------------------------------------------

#ifndef MESH_3D_SW_H
#define MESH_3D_SW_H

#include <cassert>
#include <iostream>
#include <fstream>
#include "general.H"
#include "quadrature.H"
#include "basis.H"
#include "dof.H"
#include "storage_container.H"
#include "mesh_2d_sw.H"
#include "mesh_3d.H"

//--------------------------------------------------------------------------------------
// NODE_3D_SW Class: 3D Node class for shallow-water problems
//--------------------------------------------------------------------------------------

class NODE_3D_SW : public NODE_3D
{
private:
  int tmp_id;          // Temporary id used for output
  real var[d_var_3d];  // Values of primary variables at the node

public:
  // Default constructor
  NODE_3D_SW(NODE_2D& node_xy) : NODE_3D(node_xy)
  {
    tmp_id=BAD_INDEX;
    for (int i=0; i<d_var_3d; var[i]=0., i++);
  }

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  NODE_3D_SW(std::ifstream& in) : NODE_3D(in)
  {
    tmp_id=BAD_INDEX;
    for (int i=0; i<d_var_3d; var[i]=0., i++);
  }

  // Default destructor
  ~NODE_3D_SW(){}

  // Return value: temporary id used for output, rvalue
  int TMP_ID() const {return tmp_id;}

  // Return value: temporary id used for output, lvalue
  int& TMP_ID() {return tmp_id;}

  // Return value: valued of unknowns at the node, rvalue
  real Var(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX);
    return var[VAR_3D_INDEX_INV[var_ind]];
  }

  // Return value: valued of unknowns at the node, lvalue
  real& Var(VAR var_ind)
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX);
    return var[VAR_3D_INDEX_INV[var_ind]];
  }
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ Class: Horizontal face class for shallow-water problems
//--------------------------------------------------------------------------------------

// Data shared by all horizontal faces
struct tmp_data_face_horiz
{
  // Pointers to storage for degrees of freedom
  DOF_iterator *Var_dof[d_var], *L_Var_dof[d_var], *Q_dof[d_var][d_space];  

  // Default constructor
  tmp_data_face_horiz()
  {
    for (int i=0; i<d_var; Var_dof[i]=L_Var_dof[i]=0, i++)
      for (int j=0; j<d_space; Q_dof[i][j]=0, j++);
  }
};

// Local data at each quadrature point on face
struct quad_loc_data_face_horiz
{
  // Data stored at the quadrature point
  real coord[d_space];
//   real Var[2][d_var];
// #ifdef TEST_KEY
//   real Var[d_var_3d_dt], Q[d_var_3d_dt][d_space];
// #endif
  // Index to storage for basis functions at the quadrature point
  std::size_t basis_func_ind[2][d_var_3d];
#ifdef BAROCLINIC_KEY
  // Index to storage for vertically integrated basis functions at the quadrature point
//   std::size_t basis_func_integrated_ind;
  real density_forcing;  // density forcing term
#endif

  // Default constructor
  quad_loc_data_face_horiz()
  {
    for (int j=0; j<d_space; coord[j]=0., j++);
// #ifdef TEST_KEY
//     for (int i=0; i<d_var_3d_dt; Var[i]=0., i++)
//       for (int j=0; j<d_space; Q[i][j]=0., j++);
// #endif
    for (int i=0; i<2; i++)
      for (int j=0; j<d_var_3d; basis_func_ind[i][j]=(std::size_t)BAD_INDEX, j++);
#ifdef BAROCLINIC_KEY
//     basis_func_integrated_ind=(std::size_t)BAD_INDEX;
    density_forcing=0.;
#endif
  }
};

typedef StorageContainer<tmp_data_face_horiz, empty, quad_loc_data_face_horiz> BaseStorageFaceHoriz;

class FACE_HORIZ_SW : public FACE_HORIZ, public BaseStorageFaceHoriz
{
private:
  QUADRATURE *quad;           // Pointer to the quadrature object corresponding to the face
  long_real coef_scale_d[2];  // Scaling coefficients for quadrature integration

protected:
  TRIANGLE_2D_SW* container_2d;  // Pointer to the corresponding 2D face
  char n_basis[2];               // Number of different basis functions
  char basis_index[2][d_var];    // Index of basis functions corresponding to different variables

public: 
  enum face_type {INTERIOR, SURFACE, SURFACE_STRESS, BOTTOM_NO_SLIP, BOTTOM_NO_DRAG, 
                  BOTTOM_QUAD_DRAG, BOTTOM_LIN_DRAG, BOTTOM_TEST};

  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_SW(TRIANGLE_2D_SW& elem_2d);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_SW(std::ifstream& in);

  // Default destructor
  virtual ~FACE_HORIZ_SW() {}

  // Compute mapping from the reference face
  void ComputeReferenceMapping();

  // Return value: pointer to face of the same type as the caller
  virtual FACE_HORIZ_SW* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_SW* face = new FACE_HORIZ_SW(element_2d); assert(face!=0); return face;}

  // Return value: pointer to quadrature object for the face
  QUADRATURE *Quad() const {return quad;}

  // Perform geometry update for moving free surface
  void UpdateSurfaceElevation();

//   // Return value: Is a surface face?
//   virtual bool IsSurface() const
//   {std::cerr << "File mesh_3d_sw.H: Calling unspecified function!\n"; std::exit(1); return false;}

//   // Return value: Is a bottom face?
//   virtual bool IsBottom() const
//   {std::cerr << "File mesh_3d_sw.H: Calling unspecified function!\n"; std::exit(1); return false;}

//   // Return value: Is an interior face?
//   virtual bool IsInterior() const
//   {std::cerr << "File mesh_3d_sw.H: Calling unspecified function!\n"; std::exit(1); return false;}

  // Update recursively face-element connectivity info
  // Parameters: el_ind - element index in the face; el_new - element id
  void SetElementRecursive(int el_ind, int el_new);

  // Set recursively the face (or its descendants) to be active
  // This version works for bottom faces and must be overloaded for interior and surface faces
  void SetActiveRecursive();
//   virtual void SetActiveRecursive();

  // Set recursively the face (or its descendants) to be inactive
  void UnSetActiveRecursive();

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the face in four by edge bisection
  // Parameters: node_new_ind - IDs of the new nodes, lvalue
  // Return value: id of the first child face
  virtual void Refine(int node_new_ind[vert_h]);
#endif

  // Parameters: el_ind - element index
  // Return value: scaling coefficient for quadrature integration, rvalue
  long_real coef_scale(int el_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
    return coef_scale_d[el_ind];
  }

  // Parameters: el_ind - element index
  // Return value: scaling coefficient for quadrature integration, lvalue
  long_real& coef_scale(int el_ind)
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
    return coef_scale_d[el_ind];
  }

  // Return value: matrix for "W" computation on the element below
  long_real *w_matrix() const;

  // Parameters: var_ind - global index of the unknown
  // Return value: local degrees of freedom on the coresponding 2D element for unknown "var_ind"
  DOF_iterator& Var_dof_2D(VAR var_ind) const {return container_2d->Var_dof(var_ind);}

  // Parameters: var_ind - global index of the unknown
  // Return value: local degrees of freedom on the coresponding 2D element
  //               for the right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof_2D(VAR var_ind) const {return container_2d->L_Var_dof(var_ind);}

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" for unknown "var_ind"
  DOF_iterator& Var_dof(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var && 
                 VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX && temp().Var_dof[var_ind]!=0);
    temp().Var_dof[var_ind]->SetElement(Element(el_ind));
    return *temp().Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" 
  //               for right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof(int el_ind,  VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var && 
                 VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX && temp().L_Var_dof[var_ind]!=0);
    temp().L_Var_dof[var_ind]->SetElement(Element(el_ind));
    return *temp().L_Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  //             j - coordinate index
  // Return value: local degrees of freedom on element "el_ind" 
  //               for flux vector for unknown "var_ind"
  DOF_iterator& Q_dof(int el_ind, VAR var_ind, int j) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var && j>=0 &&
                 j<d_space && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX && temp().Q_dof[var_ind][j]!=0);
    temp().Q_dof[var_ind][j]->SetElement(Element(el_ind));
    return *temp().Q_dof[var_ind][j];
  }

#ifdef BAROCLINIC_KEY
#if TURBULENCE_KEY>1
  // Parameters: el_ind - element index on the face
  // Return value: local degrees of freedom on element "el_ind" for the z-component of the density gradient
  DOF_iterator& RHO_Z_dof(int el_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && temp().Var_dof[RHO_Z]!=0);
    temp().Var_dof[RHO_Z]->SetElement(Element(el_ind));
    return *temp().Var_dof[RHO_Z];
  }
//   DOF_iterator& RO_Z_dof(int el_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && temp().Q_dof[RO][2]!=0);
//     temp().Q_dof[RO][2]->SetElement(Element(el_ind));
//     return *temp().Q_dof[RO][2];
//   }
#endif
#endif

  // ----------- Accessors: Deal with data stored at quadrature points -------------------

  // Parameters: var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the corresponding 2D quadrature point
  const real* basis_func_quad_2D(VAR var_ind) const {return container_2d->basis_func_quad(var_ind);}

  // Return value: depth at the quadrature point
  real depth_quad() const {return container_2d->depth_quad();}

  // Return value: surface elevation at the quadrature point
  real surface_elev_quad() const {return container_2d->surface_elev_quad();}

  // Return value: weight at the quadrature point
  long_real weight_quad() const {return container_2d->weight_quad();}

  // Weight for W computations at the quadrature point, rvalue
  long_real w_weight_quad() const {return weight_quad()/(depth_quad()+surface_elev_quad());}

  // Parameters: var_ind - global index of the 2D unknown
  // Return value: value of the 2D unknown at the corresponding quadrature point
  real Var_2D_quad(VAR var_ind) const {return container_2d->Var_quad(var_ind);}

#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  // Parameters: i - coordinate index
  // Return value: value of the depth intgrated velocity at the quadrature point
  real UH_quad(int i) const {return container_2d->UH_quad(i);}
#endif

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the quadrature point, rvalue
  const real* basis_func_quad(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var && 
                 basis_index[el_ind][var_ind]!=BAD_INDEX);
    return &mass_storage(quad_local().basis_func_ind[el_ind][basis_index[el_ind][var_ind]]);
  }

  // Parameters: el_ind - element index on the face; i - index of the basis function
  // Return value: pointer to values of basis functions at the quadrature point, lvalue
  real* basis_func_assign_quad(int el_ind, int i)
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && i>=0 && i<n_basis[el_ind]);
    return &mass_storage(quad_local().basis_func_ind[el_ind][i]);
  }

  // Return value: coordinates of the quadrature point, rvalue
  const real* coord_quad() const {return quad_local().coord;}

  // Return value: coordinates of the quadrature point, lvalue
  real* coord_quad() {return quad_local().coord;}

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, rvalue
  real coord_quad(int i) const
  {
    assert_debug(i>=0 && i<d_space);
    return quad_local().coord[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, lvalue
  real& coord_quad(int i)
  {
    assert_debug(i>=0 && i<d_space);
    return quad_local().coord[i];
  }

//   // Parameters: var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   virtual real Var_quad(VAR var_ind) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && IsElement(0) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Var_dof(0, var_ind)*basis_func_quad(0, var_ind);
//   }

//   // Return value: value of the "W" at the quadrature point, rvalue
//   virtual real W_quad() const
//   {
//     assert_debug(IsElement(0));
//     return -(Var_quad(U)*Normal(0)+Var_quad(V)*Normal(1))/Normal(2);
//   }

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   real Q_quad(int el_ind, VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(el_ind) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Q_dof(el_ind, var_ind, j)*basis_func_quad(el_ind, var_ind);
//   }

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   virtual real Q_quad(VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(0) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Q_dof(0, var_ind, j)*basis_func_quad(0, var_ind);
//   }

//   // Parameters: el_ind - element index; var_ind - global index of the unknown
//   // Return value: the constant component of the value of the unknown, rvalue
//   real Var_const(int el_ind, VAR var_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && 
//                  var_ind<d_var && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Var_dof(el_ind, var_ind)(0)*basis_func_quad(el_ind, var_ind)[0];
//   }

#ifdef BAROCLINIC_KEY
//   // Return value: pointer to values of vertically integrated basis functions
//   //               at the quadrature point, rvalue
//   const real* basis_func_integrated_quad() const
//   {return &mass_storage(quad_local().basis_func_integrated_ind);}

//   // Return value: pointer to values of vertically integrated basis functions
//   //               at the quadrature point, lvalue
//   real* basis_func_integrated_quad()
//   {return &mass_storage(quad_local().basis_func_integrated_ind);}

//   // Return value: depth integrated density forcing term at the quadrature point
//   real density_forcing_integrated_quad() const {return container_2d->density_forcing_quad();}

  // Return value: pointer to values of vertically integrated basis functions
  //               at the quadrature point, rvalue
  const real* basis_func_integrated_quad() const;

  // Return value: density forcing at the quadrature point
//   virtual real density_forcing_quad() const {return 0.;}
  real density_forcing_quad() const {return quad_local().density_forcing;}

  // Return value: density forcing at the quadrature point - lvalue
//   virtual real density_forcing_quad() const {return 0.;}
  real& density_forcing_quad() {return quad_local().density_forcing;}

// #if TURBULENCE_KEY>1
//   // Works for bottom faces, must be overloaded for other face types
//   // Return value: z-component of the density gradient at the quadrature point
//   virtual real density_grad_z_quad() const {return RHO_Z_dof(0)*basis_func_quad(0, RHO_Z);}
// #endif
#endif

//   // Compute flux (Neumann) boundary conditions at a quadrature point
//   // Parameters: t - time; coef - diffusion coefficients; diff - result
//   void Flux_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  virtual void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute Neumann boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  virtual void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);

  //----------------- Updaters: Update data stored at quadrature points ------------------

  // Sets the quadrature index for the corresponding 2d element
  // Parameters: ind - quadrature point index
  void SetIndex(int ind) {BaseStorageFaceHoriz::SetIndex(ind); container_2d->SetIndex(ind);}

  // Increments the quadrature index for the corresponding 2d element
  // Parameters: i - dummy parameter
  void operator++(int i) {BaseStorageFaceHoriz::operator++(i); (*container_2d)++;}

  // Initialize storage for data stored at quadrature points
  void InitializeStorage();

  // Update data shared by all horizontal faces
  // Parameters: var_new - global storage for degrees of freedom;
  //             var_flux_new - global storage for degrees of freedom for the fluxes;
  //             var_rhs_new - global storage for degrees of freedom for the right hand side
  static void Update_static(DOF var_new[d_var], DOF var_flux_new[d_var][d_space], 
                            DOF var_rhs_new[d_var]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  virtual void Update_basis_quad();

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   virtual void Update_BC_quad(real t) {}

  // Numerics
  void FaceHorizIntegration(real t);
  void FaceHorizIntegrationDiffusion(real t);
//   virtual void FaceHorizIntegrationDiffVert();
//   void FaceHorizIntegrationDiffHoriz();
  void FaceHorizIntegrationVertVelocityTop();
  void FaceHorizIntegrationVertVelocityTopMatrix();
#ifdef BAROCLINIC_KEY
//   void FaceHorizIntegrationDensity();
  void ProjectionDensityFaceHoriz();
#endif

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const FACE_HORIZ_SW&);
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_SURFACE Class: Horizontal surface face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_SURFACE : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_SURFACE(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=SURFACE; SetSurface();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_SURFACE(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_SURFACE() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_SURFACE* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_SURFACE* face = new FACE_HORIZ_SURFACE(element_2d); assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return true;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return false;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}

  // Set recursively the face (or its descendants) to be active
//   void SetActiveRecursive();

//   // Parameters: var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   real Var_quad(VAR var_ind) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && IsElement(1) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Var_dof(1, var_ind)*basis_func_quad(1, var_ind);
//   }

//   // Return value: value of the "W" at the quadrature point, rvalue
//   real W_quad() const
//   {
//     assert_debug(IsElement(1));
//     return Var_dof(1, W)*basis_func_quad(1, W);
//   }

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   real Q_quad(VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(1) &&
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
// #ifdef TEST_KEY
//     return quad_local().Q[VAR_3D_DT_INDEX_INV[var_ind]][j];
// #else
//     return 0.;
// #endif
//   }

#ifdef BAROCLINIC_KEY
  // Return value: density forcing at the quadrature point
//   real density_forcing_quad() const
//   {
//     assert_debug(IsElement(1));
//     return Var_dof(1, RO)*basis_func_integrated_quad()+
//       Var_dof(1, RO_COMPLEMENT)*basis_func_quad_2D(RO_FORCING);
//   }

// #if TURBULENCE_KEY>1
//   // Return value: z-component of the density gradient at the quadrature point
//   real density_grad_z_quad() const {return RHO_Z_dof(1)*basis_func_quad(1, RHO_Z);}
// #endif
#endif

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  void Update_basis_quad();

// #ifdef TEST_KEY
//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);
// #endif

  // Numerics
// #ifdef BAROCLINIC_KEY
//   void ProjectionDensityFaceTop();
// #endif
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_SURFACE_STRESS Class: Horizontal surface face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_SURFACE_STRESS : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_SURFACE_STRESS(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=SURFACE_STRESS; SetSurface();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_SURFACE_STRESS(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_SURFACE_STRESS() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_SURFACE_STRESS* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_SURFACE_STRESS* face = new FACE_HORIZ_SURFACE_STRESS(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return true;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return false;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}

  // Set recursively the face (or its descendants) to be active
//   void SetActiveRecursive();

// #ifdef BAROCLINIC_KEY
// #if TURBULENCE_KEY>1
//   // Return value: z-component of the density gradient at the quadrature point
//   real density_grad_z_quad() const {return RHO_Z_dof(1)*basis_func_quad(1, RHO_Z);}
// #endif
// #endif

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  void Update_basis_quad();
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_INTERIOR Class: Horizontal interior face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_INTERIOR : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_INTERIOR(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=INTERIOR; SetInterior();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_INTERIOR(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_INTERIOR() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_INTERIOR* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_INTERIOR* face = new FACE_HORIZ_INTERIOR(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return false;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return false;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return true;}

  // Set recursively the face (or its descendants) to be active
//   void SetActiveRecursive();

//   // Parameters: var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   real Var_quad(VAR var_ind) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && IsElement(0) && IsElement(1) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return 0.5*(Var_dof(0, var_ind)*basis_func_quad(0, var_ind)+
//                 Var_dof(1, var_ind)*basis_func_quad(1, var_ind));
//   }

//   // Return value: value of the "W" at the quadrature point, rvalue
//   real W_quad() const
//   {
//     assert_debug(IsElement(1));
//     return Var_dof(1, W)*basis_func_quad(1, W);
//   }

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   real Q_quad(VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(0) && 
//                  IsElement(1) && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return 0.5*(Q_dof(0, var_ind, j)*basis_func_quad(0, var_ind)+
//                 Q_dof(1, var_ind, j)*basis_func_quad(1, var_ind));
//   }

#ifdef BAROCLINIC_KEY
  // Return value: density forcing at the quadrature point
//   real density_forcing_quad() const {return Var_dof(1, RO)*basis_func_integrated_quad()+
//       Var_dof(1, RO_COMPLEMENT)*basis_func_quad_2D(RO_FORCING);}

// #if TURBULENCE_KEY>1
//   // Return value: z-component of the density gradient at the quadrature point
//   real density_grad_z_quad() const
//   {return 0.5*(RHO_Z_dof(0)*basis_func_quad(0, RHO_Z)+RHO_Z_dof(1)*basis_func_quad(1, RHO_Z));}
// #endif
#endif

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  void Update_basis_quad();
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_BOTTOM_NO_SLIP Class: Bottom no-slip face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_BOTTOM_NO_SLIP : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_BOTTOM_NO_SLIP(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=BOTTOM_NO_SLIP; SetBottom();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_BOTTOM_NO_SLIP(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_BOTTOM_NO_SLIP() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_BOTTOM_NO_SLIP* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_BOTTOM_NO_SLIP* face = new FACE_HORIZ_BOTTOM_NO_SLIP(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return false;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return true;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}

//   // Parameters: var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   real Var_quad(VAR var_ind) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && IsElement(0) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return (var_ind==U || var_ind==V) ? 0. : Var_dof(0, var_ind)*basis_func_quad(0, var_ind);
//   }

//   // Return value: value of the "W" at the quadrature point, rvalue
//   real W_quad() const {return 0.;}

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_BOTTOM_NO_DRAG Class: Bottom no-drag face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_BOTTOM_NO_DRAG : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_BOTTOM_NO_DRAG(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=BOTTOM_NO_DRAG; SetBottom();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_BOTTOM_NO_DRAG(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_BOTTOM_NO_DRAG() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_BOTTOM_NO_DRAG* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_BOTTOM_NO_DRAG* face = new FACE_HORIZ_BOTTOM_NO_DRAG(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return false;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return true;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}


  // Parameters: var_ind - global index of the unknown; j - coordinate index
  // Return value: value of the flux of unknown at the quadrature point, rvalue
  real Q_quad(VAR var_ind, int j) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(0) && 
                 VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
    return 0.;
  }

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_BOTTOM_QUAD_DRAG Class: Bottom quadratic drag face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_BOTTOM_QUAD_DRAG : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_BOTTOM_QUAD_DRAG(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=BOTTOM_QUAD_DRAG; SetBottom();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_BOTTOM_QUAD_DRAG(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_BOTTOM_QUAD_DRAG() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_BOTTOM_QUAD_DRAG* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_BOTTOM_QUAD_DRAG* face = new FACE_HORIZ_BOTTOM_QUAD_DRAG(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return false;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return true;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}

  // Numerics
//   void FaceHorizIntegrationDiffVert();

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_BOTTOM_LIN_DRAG Class: Bottom linear drag face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_HORIZ_BOTTOM_LIN_DRAG : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_BOTTOM_LIN_DRAG(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=BOTTOM_LIN_DRAG; SetBottom();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_BOTTOM_LIN_DRAG(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_BOTTOM_LIN_DRAG() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_BOTTOM_LIN_DRAG* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_BOTTOM_LIN_DRAG* face = new FACE_HORIZ_BOTTOM_LIN_DRAG(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return false;}

//   // Return value: Is an interior face?
//   bool IsBottom() const {return true;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}

  // Numerics
//   void FaceHorizIntegrationDiffVert();

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};



//--------------------------------------------------------------------------------------
//  FACE_HORIZ_BOTTOM_TEST Class: Bottom test face class for shallow-water problems.
//                                Used for test problems with analytical solution
//--------------------------------------------------------------------------------------

#ifdef TEST_KEY
class FACE_HORIZ_BOTTOM_TEST : public FACE_HORIZ_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  FACE_HORIZ_BOTTOM_TEST(TRIANGLE_2D_SW& element_2d) : FACE_HORIZ_SW(element_2d)
  {Type()=BOTTOM_TEST; SetBottom();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_HORIZ_BOTTOM_TEST(std::ifstream& in) : FACE_HORIZ_SW(in) {}

  // Default destructor
  ~FACE_HORIZ_BOTTOM_TEST() {}

  // Return value: pointer to face of the same type as the caller
  FACE_HORIZ_BOTTOM_TEST* CreateFaceHoriz(TRIANGLE_2D_SW& element_2d)
  {FACE_HORIZ_BOTTOM_TEST* face = new FACE_HORIZ_BOTTOM_TEST(element_2d);
    assert(face!=0); return face;}

//   // Return value: Is a surface face?
//   bool IsSurface() const {return false;}

//   // Return value: Is a bottom face?
//   bool IsBottom() const {return true;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return false;}

//   // Parameters: var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   real Var_quad(VAR var_ind) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && IsElement(0) &&
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return quad_local().Var[VAR_3D_INDEX_INV[var_ind]];
//   }

//   // Return value: value of the "W" at the quadrature point, rvalue
//   real W_quad() const
//   {
//     assert_debug(IsElement(0));
//     return quad_local().Var[VAR_3D_INDEX_INV[W]];
//   }

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};
#endif



//--------------------------------------------------------------------------------------
//  FACE_VERT Class: Vertical face class for shallow-water problems
//--------------------------------------------------------------------------------------

// Data shared by all vertical faces
struct tmp_data_face_vert
{
  // Pointers to storage for degrees of freedom
  DOF_iterator *Var_dof[d_var], *L_Var_dof[d_var], *Q_dof[d_var][d_space];  

  // Default constructor
  tmp_data_face_vert()
  {
    for (int i=0; i<d_var; Var_dof[i]=L_Var_dof[i]=0, i++)
      for (int j=0; j<d_space; Q_dof[i][j]=0, j++);
  }
};

// Local data at each quadrature point on face
struct quad_loc_data_face_vert
{
  // Data stored at the quadrature point
//   real Var[2][d_var_3d_dt], coord[d_space];
  real coord[d_space];
  long_real weight;
// #ifdef TEST_KEY
//   real Q[d_var_3d_dt][d_space];
// #endif
  // Index to storage for basis functions at the quadrature point
  std::size_t basis_func_ind[2][d_var_3d];
// #ifdef BAROCLINIC_KEY
//   // Index to storage for vertically integrated basis functions at the quadrature point
//   std::size_t basis_func_integrated_ind[2];
// #endif

  // Default constructor
  quad_loc_data_face_vert() 
  {
    for (int j=0; j<d_space; coord[j]=0., j++);
    weight=0.;
// #ifdef TEST_KEY
//     for (int i=0; i<d_var_3d_dt; i++)
//       for (int j=0; j<d_space; Q[i][j]=0., j++);
// #endif
    for (int i=0; i<2; i++)
    {
//       for (int j=0; j<d_var_3d_dt; Var[i][j]=0., j++);
      for (int j=0; j<d_var_3d; basis_func_ind[i][j]=(std::size_t)BAD_INDEX, j++);
// #ifdef BAROCLINIC_KEY
//       basis_func_integrated_ind[i]=(std::size_t)BAD_INDEX;
// #endif
    }
  }
};

typedef StorageContainer<tmp_data_face_vert, empty, quad_loc_data_face_vert> BaseStorageFaceVert;

class FACE_VERT_SW : public FACE_VERT, public BaseStorageFaceVert
{
private:
  QUADRATURE *quad;            // Pointer to the quadrature object corresponding to the face
  long_real coef_scale_d[2];   // Scaling coefficients for quadrature integration

protected:
  EDGE_2D_SW* container_2d;    // Pointer to the corresponding 2D edge
  char n_basis[2];             // Number of different basis functions
  char basis_index[2][d_var];  // Index of basis functions corresponding to different variables
  // Pointer to the Riemann solver function
//   void (*RiemannSolver)(const long_real[d_space], const real[d_var], const real[d_var], long_real[d_var]);

public: 
  // Default constructor
  // Parameters: edge - corresponding 2D edge
  FACE_VERT_SW(EDGE_2D_SW& edge);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_SW(std::ifstream& in);

  // Default destructor
  virtual ~FACE_VERT_SW() {}

  // Return value: pointer to face of the same type as the caller
  virtual FACE_VERT_SW* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_SW* face = new FACE_VERT_SW(edge); assert(face!=0); return face;}

  // Compute mapping from the reference face
  void ComputeReferenceMapping();

  // Return value: pointer to quadrature object for the face
  QUADRATURE *Quad() const {return quad;}

  // Perform geometry update for moving free surface
  void UpdateSurfaceElevation();

//   // Return value: Is an interior face?
//   // works for boundary faces, must be overloaded for interior faces
//   virtual bool IsInterior() const {return false;}

  // Update recursively face-element connectivity info
  // Parameters: el_ind - element index in the face; el_new - element id
  void SetElementRecursive(int el_ind, int el_new);

  // Set recursively the face (or its descendants) to be active
  // This version works for boundary faces and must be overloaded for interior faces
  void SetActiveRecursive();
//   virtual void SetActiveRecursive();

  // Set recursively the face (or its descendants) to be inactive
  void UnSetActiveRecursive();

  // Set recursively the face (or its descendants) to be updated
  void UnSetUpdatedRecursive();

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Parameters: el_ind - element index; k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, rvalue
  real JumpBoundary(int el_ind, int k) const;

  // Parameters: el_ind - element index; k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, lvalue
  real& JumpBoundary(int el_ind, int k);

  // Parameters: el_ind - element index; k - 0=inflow, 1=outflow
  // Return value: size of the jump in the unknown used for adaptivity over
  //               the inflow/outflow boundary, rvalue
  real Jump(int el_ind, int k) const;

  // Parameters: el_ind - element index; k - 0=inflow, 1=outflow
  // Return value: size of the jump in the unknown used for adaptivity over
  //               the inflow/outflow boundary, rvalue
  real& Jump(int el_ind, int k);

  // Refine the face horizontally
  // Parameters: node_new_ind - IDs of the new nodes, lvalue
  // Return value: id of the first child face
  virtual void RefineHoriz(int node_new_ind[2]);
#endif

  // Parameters: el_ind - element index
  // Return value: scaling coefficient for quadrature integration, rvalue
  long_real coef_scale(int el_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
    return coef_scale_d[el_ind];
  }

  // Parameters: el_ind - element index
  // Return value: scaling coefficient for quadrature integration, lvalue
  long_real& coef_scale(int el_ind)
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
    return coef_scale_d[el_ind];
  }

  // Parameters: el_ind - element index
  // Return value: 2D scaling coefficient for quadrature integration
  long_real coef_scale_elev(int el_ind) const {return container_2d->coef_scale(el_ind);}

  // Parameters: el_ind - element index; var_ind - global index of the unknown
  // Return value: local degrees of freedom on the coresponding 2D element for unknown "var_ind"
  DOF_iterator& Var_dof_2D(int el_ind, VAR var_ind) const
  {return container_2d->Var_dof(el_ind, var_ind);}

  // Parameters: el_ind - element index; var_ind - global index of the unknown
  // Return value: local degrees of freedom on the coresponding 2D element
  //               for the right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof_2D(int el_ind, VAR var_ind) const
  {return container_2d->L_Var_dof(el_ind, var_ind);}

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" for unknown "var_ind"
  DOF_iterator& Var_dof(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var &&
                 VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX && temp().Var_dof[var_ind]!=0);
    temp().Var_dof[var_ind]->SetElement(Element(el_ind));
    return *temp().Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" 
  //               for right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var &&
                 VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX && temp().L_Var_dof[var_ind]!=0);
    temp().L_Var_dof[var_ind]->SetElement(Element(el_ind));
    return *temp().L_Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  //             j - coordinate index
  // Return value: local degrees of freedom on element "el_ind" 
  //               for flux vector for unknown "var_ind"
  DOF_iterator& Q_dof(int el_ind, VAR var_ind, int j) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var && j>=0 &&
                 j<d_space && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX && temp().Q_dof[var_ind][j]!=0);
    temp().Q_dof[var_ind][j]->SetElement(Element(el_ind));
    return *temp().Q_dof[var_ind][j];
  }

  // ----------- Accessors: Deal with data stored at quadrature points -------------------

  // Parameters: var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the corresponding 2D quadrature point
  const real* basis_func_quad_2D(int el_ind, VAR var_ind) const
  {return container_2d->basis_func_quad(el_ind, var_ind);}

  // Return value: depth at the quadrature point
  real depth_quad() const {return container_2d->depth_quad();}

  // Return value: surface elevation at the quadrature point
  real surface_elev_quad() const {return container_2d->surface_elev_quad();}

  // Return value: weight at the quadrature point, rvalue
  long_real weight_quad() const {return quad_local().weight;};

  // Return value: weight at the quadrature point, lvalue
  long_real& weight_quad() {return quad_local().weight;};

  // Return value: weight at the quadrature point used for 2D terms
  long_real elev_weight_quad() const {return quad_local().weight/(depth_quad()+surface_elev_quad());}

  // Parameters: el_ind - element index; var_ind - global index of the 2D unknown
  // Return value: value of the 2D unknown at the corresponding quadrature point
  real Var_2D_quad(int el_ind, VAR var_ind) const {return container_2d->Var_quad(el_ind, var_ind);}

#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  // Parameters: i - coordinate index
  // Return value: value of the depth integrated velocity at the quadrature point
  real UH_quad(int i) const {return container_2d->UH_quad(i);}
#endif

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the quadrature point, rvalue
  const real* basis_func_quad(int el_ind, VAR var_ind) const
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var &&
                 basis_index[el_ind][var_ind]!=BAD_INDEX);
    return &mass_storage(quad_local().basis_func_ind[el_ind][basis_index[el_ind][var_ind]]);
  }

  // Parameters: el_ind - element index on the face; i - index of the basis function
  // Return value: pointer to values of basis functions at the quadrature point, lvalue
  real* basis_func_assign_quad(int el_ind, int i)
  {
    assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && i>=0 && i<n_basis[el_ind]);
    return &mass_storage(quad_local().basis_func_ind[el_ind][i]);
  }

  // Return value: coordinates of the quadrature point, rvalue
  const real* coord_quad() const {return quad_local().coord;}

  // Return value: coordinates of the quadrature point, lvalue
  real* coord_quad() {return quad_local().coord;}

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, rvalue
  real coord_quad(int i) const
  {
    assert_debug(i>=0 && i<d_space);
    return quad_local().coord[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, lvalue
  real& coord_quad(int i)
  {
    assert_debug(i>=0 && i<d_space);
    return quad_local().coord[i];
  }

//   // Parameters: var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   virtual real Var_quad(int el_ind, VAR var_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && var_ind>=0 && var_ind<d_var &&
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return quad_local().Var[el_ind][VAR_3D_DT_INDEX_INV[var_ind]];
//   }

//   // Parameters: el_ind - element index
//   // Return value: value of the "W" at the quadrature point, rvalue
//   real W_quad(int el_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
//     return Var_dof(el_ind, W)*basis_func_quad(el_ind, W);
//   }

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   virtual real Q_quad(VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(0) && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return 0.;
//   }

//   // Parameters: el_ind - element index; var_ind - global index of the unknown
//   // Return value: the constant component of the value of the unknown, rvalue
//   real Var_const(int el_ind, VAR var_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && 
//                  var_ind<d_var && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Var_dof(el_ind, var_ind)(0)*basis_func_quad(el_ind, var_ind)[0];
//   }

#ifdef BAROCLINIC_KEY
//   // Parameters: el_ind - element index on the face
//   // Return value: pointer to values of vertically integrated basis functions
//   //               at the quadrature point, rvalue
//   const real* basis_func_integrated_quad(int el_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
//     return &mass_storage(quad_local().basis_func_integrated_ind[el_ind]);
//   }

//   // Parameters: el_ind - element index on the face
//   // Return value: pointer to values of vertically integrated basis functions
//   //               at the quadrature point, lvalue
//   real* basis_func_integrated_quad(int el_ind)
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind));
//     return &mass_storage(quad_local().basis_func_integrated_ind[el_ind]);
//   }

//   // Return value: depth integrated density forcing term at the quadrature point
//   real density_forcing_integrated_quad() const {return container_2d->density_forcing_quad();}

  // Parameters: el_ind - element index on the face
  // Return value: pointer to values of vertically integrated basis functions
  //               at the quadrature point, rvalue
  const real* basis_func_integrated_quad(int el_ind) const;

//   // Return value: density forcing at the quadrature point
//   // Works for boundary faces, must be overloaded for interior faces
//   virtual real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0));
//     return Var_dof(0, RHO)*basis_func_integrated_quad(0)+Var_dof(0, RHO_FRS)*basis_func_quad_2D(0, XI);
//   }

//   virtual real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0));
//     return Var_dof(0, RHO_FRS)*basis_func_quad(0, RHO_FRS);
//   }
//   virtual real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0));
//     return Var_dof(0, RO)*basis_func_integrated_quad(0)+
//       Var_dof(0, RO_COMPLEMENT)*basis_func_quad_2D(0, RO_FORCING);
//   }
//   real density_forcing_quad() const;
  // Return value: density forcing at the quadrature point
  real density_forcing_quad() const
  {
    assert_debug(IsElement(0));
    if (IsInterior())
      return 0.5*(Var_dof(0, RHO)*basis_func_integrated_quad(0)+
                  Var_dof(0, RHO_FRS)*basis_func_quad_2D(0, XI)+
                  Var_dof(1, RHO)*basis_func_integrated_quad(1)+
                  Var_dof(1, RHO_FRS)*basis_func_quad_2D(1, XI));
    else
      return Var_dof(0, RHO)*basis_func_integrated_quad(0)+
        Var_dof(0, RHO_FRS)*basis_func_quad_2D(0, XI);
  }
#endif

//   // Compute flux (Neumann) boundary conditions at a quadrature point
//   // Parameters: t - time; coef - diffusion coefficients; diff - result
//   void Flux_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  virtual void BC_Dirichlet_quad(real t, const real param[], real var[d_var])
  {std::cerr << "File mesh_3d_sw.H: Dirichlet boundary condition is not spacified for this face type!\n";
    std::exit(1);}

  // Compute Neumann boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  virtual void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);

  //----------------- Updaters: Update data stored at quadrature points ------------------

  // Sets the quadrature index for the corresponding 2d edge
  // Parameters: ind - quadrature point index
  void SetIndex(int ind)
  {BaseStorageFaceVert::SetIndex(ind); container_2d->SetIndex(Quad()->GetSubIndex(0, ind));}

  // Increments the quadrature index for the corresponding 2d edge
  // Parameters: i - dummy parameter
  void operator++(int i)
  {BaseStorageFaceVert::operator++(i); container_2d->SetIndex(Quad()->GetSubIndex(0, GetIndex()));}

  // Initialize storage for data stored at quadrature points
  void InitializeStorage();

  // Update data shared by all vertical faces
  // Parameters: var_new - global storage for degrees of freedom;
  //             var_flux_new - global storage for degrees of freedom for the fluxes;
  //             var_rhs_new - global storage for degrees of freedom for the right hand side
  static void Update_static(DOF var_new[d_var], DOF var_flux_new[d_var][d_space], 
                            DOF var_rhs_new[d_var]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  virtual void Update_basis_quad();

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_variables_quad(real t);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   virtual void Update_BC_quad(real t);

  // Numerics
  void FaceVertIntegration(real t);
  void FaceVertIntegrationDiffusion(real t);
//   void FaceVertIntegrationDiffHoriz();
// #ifdef BAROCLINIC_KEY
//   void FaceVertIntegrationDensity();
// #endif

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const FACE_VERT_SW&);
};



//--------------------------------------------------------------------------------------
//  FACE_VERT_INTERIOR Class: Vertical interior face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_VERT_INTERIOR : public FACE_VERT_SW
{
public:
  // Default constructor
  // Parameters: edge_2d - corresponding 2D edge
  FACE_VERT_INTERIOR(EDGE_INTERIOR& edge_2d) : FACE_VERT_SW(edge_2d) {SetInterior();}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_INTERIOR(std::ifstream& in) : FACE_VERT_SW(in) {}

  // Default destructor
  ~FACE_VERT_INTERIOR() {}

  // Return value: pointer to face of the same type as the caller
  FACE_VERT_INTERIOR* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_INTERIOR* face = new FACE_VERT_INTERIOR((EDGE_INTERIOR&)edge); 
    assert(face!=0); return face;}

//   // Return value: Is an interior face?
//   bool IsInterior() const {return true;}

  // Set recursively the face (or its descendants) to be active
//   void SetActiveRecursive();

//   // Parameters: el_ind - element index; var_ind - global index of the unknown
//   // Return value: value of the unknown at the quadrature point, rvalue
//   real Var_quad(int el_ind, VAR var_ind) const
//   {
//     assert_debug(el_ind>=0 && el_ind<2 && IsElement(el_ind) && var_ind>=0 && var_ind<d_var && 
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Var_dof(el_ind, var_ind)*basis_func_quad(el_ind, var_ind);
//   }

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   real Q_quad(VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(0) && 
//                  IsElement(1) && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return 0.5*(Q_dof(0, var_ind, j)*basis_func_quad(0, var_ind)+
//                 Q_dof(1, var_ind, j)*basis_func_quad(1, var_ind));
//   }
#ifdef BAROCLINIC_KEY
//   // Return value: density forcing at the quadrature point
//   real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0) && IsElement(1));
//     return 0.5*(Var_dof(0, RHO)*basis_func_integrated_quad(0)+
//                 Var_dof(0, RHO_FRS)*basis_func_quad_2D(0, XI)+
//                 Var_dof(1, RHO)*basis_func_integrated_quad(1)+
//                 Var_dof(1, RHO_FRS)*basis_func_quad_2D(1, XI));
//   }

//   real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0) && IsElement(1));
//     return 0.5*(Var_dof(0, RHO_FRS)*basis_func_quad(0, RHO_FRS)+
//                 Var_dof(1, RHO_FRS)*basis_func_quad(1, RHO_FRS));
//   }
//   // Return value: density forcing at the quadrature point
//   real density_forcing_quad() const
//   {
//     assert_debug(IsElement(0) && IsElement(1));
//     return 0.5*(Var_dof(0, RO)*basis_func_integrated_quad(0)+
//                 Var_dof(0, RO_COMPLEMENT)*basis_func_quad_2D(0, RO_FORCING)+
//                 Var_dof(1, RO)*basis_func_integrated_quad(1)+
//                 Var_dof(1, RO_COMPLEMENT)*basis_func_quad_2D(1, RO_FORCING));
//   }
#endif
};



//--------------------------------------------------------------------------------------
//  FACE_VERT_LAND Class: Vertical land face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_VERT_LAND : public FACE_VERT_SW
{
public:
  // Default constructor
  // Parameters: edge_2d - corresponding 2D edge
  FACE_VERT_LAND(EDGE_LAND& edge_2d) : FACE_VERT_SW(edge_2d) {}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_LAND(std::ifstream& in) : FACE_VERT_SW(in) {}

  // Default destructor
  ~FACE_VERT_LAND() {}

  // Return value: pointer to face of the same type as the caller
  FACE_VERT_LAND* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_LAND* face = new FACE_VERT_LAND((EDGE_LAND&)edge); assert(face!=0); return face;}

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);
};



//--------------------------------------------------------------------------------------
//  FACE_VERT_SEA Class: Vertical sea face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_VERT_SEA : public FACE_VERT_SW
{
public:
  // Default constructor
  // Parameters: edge_2d - corresponding 2D edge
  FACE_VERT_SEA(EDGE_SEA& edge_2d) : FACE_VERT_SW(edge_2d) {}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_SEA(std::ifstream& in) : FACE_VERT_SW(in) {}

  // Default destructor
  ~FACE_VERT_SEA() {}

  // Return value: pointer to face of the same type as the caller
  FACE_VERT_SEA* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_SEA* face = new FACE_VERT_SEA((EDGE_SEA&)edge); assert(face!=0); return face;}

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);

  // Numerics
  //   void FaceVertIntegration();

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);
};



//--------------------------------------------------------------------------------------
//  FACE_VERT_RIVER Class: Vertical river face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_VERT_RIVER : public FACE_VERT_SW
{
public:
  // Default constructor
  // Parameters: edge_2d - corresponding 2D edge
  FACE_VERT_RIVER(EDGE_RIVER& edge_2d) : FACE_VERT_SW(edge_2d) {}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_RIVER(std::ifstream& in) : FACE_VERT_SW(in) {}

  // Default destructor
  ~FACE_VERT_RIVER() {}

  // Return value: pointer to face of the same type as the caller
  FACE_VERT_RIVER* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_RIVER* face = new FACE_VERT_RIVER((EDGE_RIVER&)edge); assert(face!=0); return face;}

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};



//--------------------------------------------------------------------------------------
//  FACE_VERT_RADIATION Class: Vertical radiation face class for shallow-water problems
//--------------------------------------------------------------------------------------

class FACE_VERT_RADIATION : public FACE_VERT_SW
{
public:
  // Default constructor
  // Parameters: edge_2d - corresponding 2D edge
  FACE_VERT_RADIATION(EDGE_RADIATION& edge_2d) : FACE_VERT_SW(edge_2d) {}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_RADIATION(std::ifstream& in) : FACE_VERT_SW(in) {}

  // Default destructor
  ~FACE_VERT_RADIATION() {}

  // Return value: pointer to face of the same type as the caller
  FACE_VERT_RADIATION* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_RADIATION* face = new FACE_VERT_RADIATION((EDGE_RADIATION&)edge); 
    assert(face!=0); return face;}

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);
};



#ifdef TEST_KEY
//--------------------------------------------------------------------------------------
//  FACE_VERT_UNKNOWN Class: Vertical unknown(Dirichlet) face class for shallow-water problems
//                           Used in test problems with analytical solution
//--------------------------------------------------------------------------------------

class FACE_VERT_UNKNOWN : public FACE_VERT_SW
{
public:
  // Default constructor
  // Parameters: edge_2d - corresponding 2D edge
  FACE_VERT_UNKNOWN(EDGE_UNKNOWN& edge_2d) : FACE_VERT_SW(edge_2d) {}

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  FACE_VERT_UNKNOWN(std::ifstream& in) : FACE_VERT_SW(in) {}

  // Default destructor
  ~FACE_VERT_UNKNOWN() {}

  // Return value: pointer to face of the same type as the caller
  FACE_VERT_UNKNOWN* CreateFaceVert(EDGE_2D_SW& edge)
  {FACE_VERT_UNKNOWN* face = new FACE_VERT_UNKNOWN((EDGE_UNKNOWN&)edge); 
    assert(face!=0); return face;}

//   // Parameters: var_ind - global index of the unknown; j - coordinate index
//   // Return value: value of the flux of unknown at the quadrature point, rvalue
//   real Q_quad(VAR var_ind, int j) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && IsElement(0) &&
//                  VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
//     return quad_local().Q[VAR_3D_DT_INDEX_INV[var_ind]][j];
//   }

//   // Update often changed data at quadrature points (unknowns)
//   // Parameters: t - time
//   void Update_BC_quad(real t);

  // Compute boundary conditions for primary variables
  // Parameters: t - time; param - parameter list; var - result
  void BC_Dirichlet_quad(real t, const real param[], real var[d_var]);

  // Compute flux (Neumann) boundary conditions at a quadrature point
  // Parameters: t - time; coef - diffusion coefficients; diff - result
  void BC_Neumann_quad(real t, const real coef[d_var][d_space], real diff[d_var][d_space]);
};
#endif



//--------------------------------------------------------------------------------------
// PRISM_SW Class: Element class in 3D for shallow-water problems
//--------------------------------------------------------------------------------------

// Data shared by all prisms
struct tmp_data_prism
{
  // Field used for norm and error computation
  long_real norm[d_var_3d];
  // Pointers to storage for degrees of freedom
  DOF_iterator *Var_dof[d_var], *L_Var_dof[d_var], *Q_dof[d_var][d_space];
#ifdef MASS_CONSERVATION_KEY // Variables for mass conservation computation
  long_real mass, mass_change, mass_stage[max_stages];
#endif

  // Default constructor
  tmp_data_prism()
  {
    for (int i=0; i<d_var_3d; norm[i]=0., i++);
    for (int i=0; i<d_var; Var_dof[i]=L_Var_dof[i]=0, i++)
      for (int j=0; j<d_space; Q_dof[i][j]=0, j++);
#ifdef MASS_CONSERVATION_KEY
    mass=0.;
#endif
  }
};

// Data at quadrature points shared by all prisms
struct quad_tmp_data_prism
{
  // Index to storage for basis functions at the quadrature points for parallel prisms
  std::size_t basis_func_ind[d_var_3d];

  // Default constructor
  quad_tmp_data_prism()
  {for (int i=0; i<d_var_3d; basis_func_ind[i]=(std::size_t)BAD_INDEX, i++);}
};

// Local data at each quadrature point on prism
struct quad_loc_data_prism
{
  // Data stored at the quadrature point
  real coord[d_space];
  long_real weight;
  // Index to storage for basis sunctions and their gradients at the quadrature point
  std::size_t basis_func_ind[d_var_3d];
  std::size_t basis_grad_ind[d_var_3d][d_space];
// #ifdef BAROCLINIC_KEY
//   // Index to storage for vertically integrated basis functions at the quadrature point
//   std::size_t basis_func_integrated_ind;
// #endif

  // Default constructor
  quad_loc_data_prism()
  {
    for (int j=0; j<d_space; coord[j]=0., j++);
    weight=0.;
    for (int i=0; i<d_var_3d; i++)
    {
      basis_func_ind[i]=(std::size_t)BAD_INDEX;
      for (int j=0; j<d_space; basis_grad_ind[i][j]=(std::size_t)BAD_INDEX, j++);
    }
// #ifdef BAROCLINIC_KEY
//     basis_func_integrated_ind=(std::size_t)BAD_INDEX;
// #endif
  }
};

typedef StorageContainer<tmp_data_prism, quad_tmp_data_prism, quad_loc_data_prism> BaseStoragePrism;

class PRISM_SW : public PRISM, public BaseStoragePrism
{
private:
  unsigned char order;           // Order of the approximating space
  QUADRATURE *quad;              // Pointer to quadrature object corresponding to the element
  QUADRATURE *vert;              // Pointer to mapping to the vertices of the element
  std::size_t w_matrix_ind;      // Index to storage for matrix for "W" computation
#ifdef ADAPTIVITY_H_3D_KEY
  real min, max, abs_max;
  real s_jump[2], s_boundary[2]; // Variables for shock-detector
#endif

protected:
  BASIS *basis[d_var_3d];        // Pointers to the objects describing the basis functions
  char n_basis;                  // Number of different basis functions
  char basis_index[d_var];       // Index of basis functions corresponding to different variables
  TRIANGLE_2D_SW* container_2d;  // Pointer to the corresponding 2D face

public:
  enum prism_type {PARALLEL, NONPARALLEL};

  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  PRISM_SW(TRIANGLE_2D_SW& elem_2d);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  PRISM_SW(std::ifstream& in);

  // Default destructor
  virtual ~PRISM_SW() {}

  // Compute mapping from the reference prism
  virtual void ComputeReferenceMapping();

  // Return value: pointer to a prism of the same type as the caller
  virtual PRISM_SW* CreatePrism(TRIANGLE_2D_SW& face)
  {PRISM_SW* prism = new PRISM_SW(face); assert(prism!=0); return prism;}

  // Return value: order of the approximation on the element, rvalue
  int Order() const {return order;}

  // Return value: pointer to quadrature object for the element
  QUADRATURE *Quad() const {return quad;}

  // Return value: pointer to quadrature object for the vertices
  QUADRATURE *Vert() const {return vert;}

  // Parameters: i - basis index
  // Return value: pointer to basis object, rvalue
  const BASIS *Basis(int i) const
  {
    assert_debug(i>=0 && i<d_var_3d && basis[i]!=0);
    return basis[i];
  }

  // Parameters: var_ind - variable index
  // Return value: pointer to basis object, rvalue
  const BASIS *Basis_dof(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && basis_index[var_ind]!=BAD_INDEX);
    return basis[basis_index[var_ind]];
  }

  // Return value: number of different basis functions
  char NBasis() const {return n_basis;}

  // Parameters: var_ind - variable index
  // Return value: index of basis object corresponding to the unknown
  char Basis_index(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var);
    return basis_index[var_ind];
  }  

  // Perform geometry update for moving free surface
  void UpdateSurfaceElevation();

  // Return value: is a parallel (interior) prism?
  bool IsParallel() const {return Type()==PARALLEL;}

//   // Return value: is a surface prism?
//   bool IsSurface() const {return Type()==SURFACE || Type()==SURFACE_AND_BOTTOM;}

//   // Return value: is a bottom prism?
//   bool IsBottom() const {return Type()==BOTTOM || Type()==SURFACE_AND_BOTTOM;}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the element
  // Return value: id of the first child element
  virtual void Refine();

  // Unrefine the element
  virtual void UnRefine();
#endif

#ifdef ADAPTIVITY_H_2D_KEY

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, rvalue
  real JumpBoundary(int k) const {return container_2d->JumpBoundary(k);}

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, lvalue
  real& JumpBoundary(int k) {return container_2d->JumpBoundary(k);}

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: jump in the unknown used for adaptivity on inflow/outflow boundary, rvalue
  real Jump(int k) const {return container_2d->Jump(k);}

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: jump in the unknown used for adaptivity on inflow/outflow boundary, lvalue
  real& Jump(int k) {return container_2d->Jump(k);}
#endif

#ifdef ADAPTIVITY_H_3D_KEY
  // Initialize data used for shock-detector
  void InitializeAdaptivityData();

  // Return value: maximum value of the unknown used for adaptivity on the element
  real GetMax() const {return max;}

  // Return value: minimum value of the unknown used for adaptivity on the element
  real GetMin() const {return min;}

  // Return value: maximum absolute value of the unknown used for adaptivity on the element
  real GetMaxAbs() const {return abs_max;}

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, rvalue
  real JumpBoundary(int k) const
  {
    assert_debug(k>=0 && k<2);
    return s_boundary[k];
  }

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: size of the inflow/outflow boundary, lvalue
  real& JumpBoundary(int k)
  {
    assert_debug(k>=0 && k<2);
    return s_boundary[k];
  }

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: jump in the unknown used for adaptivity on inflow/outflow boundary, rvalue
  real Jump(int k) const
  {
    assert_debug(k>=0 && k<2);
    return s_jump[k];
  }

  // Parameters: k - 0=inflow, 1=outflow
  // Return value: jump in the unknown used for adaptivity on inflow/outflow boundary, lvalue
  real& Jump(int k)
  {
    assert_debug(k>=0 && k<2);
    return s_jump[k];
  }
#endif

  // Return value: scaling coefficient for quadrature integration, rvalue
  long_real scale() const {return Mapping().GetScale();}

  // Return value: value in the "norm" field, lvalue
  static long_real& norm(int i) 
  {
    assert_debug(i>=0 && i<d_var_3d);
    return temp_storage.norm[i];
  }

#ifdef MASS_CONSERVATION_KEY
  // Return value: value in the "mass" field, lvalue
  static long_real& mass() {return temp_storage.mass;}

  // Return value: value in the "mass_change" field, lvalue
  static long_real& mass_change() {return temp_storage.mass_change;}

  // Return value: value in the "mass_stage[i]" field, lvalue
  static long_real& mass_stage(int i)
  {
    assert_debug(i>=0 && i<d_var_3d);
    return temp_storage.mass_stage[i];
  }
#endif

  // Parameters: var_ind - global index of the unknown; j - node index
  // Return value: pointer to values of basis functions at the nodes, rvalue
  const real* basis_func_node(VAR var_ind, int j, func_type basis_func) const 
  {
    assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<2*vert_h &&
                 basis_index[var_ind]!=BAD_INDEX);
    Basis_dof(var_ind)->basis_func_ref(Vert()->p(j), Mapping(), basis_func);
    return basis_func;
  }

  // Return value: pointer to matrix for "W" compuutation, rvalue
  const long_real* w_matrix() const {return (long_real*)&mass_storage(w_matrix_ind);}

  // Return value: pointer to matrix for "W" compuutation, lvalue
  long_real* w_matrix() {return (long_real*)&mass_storage(w_matrix_ind);}

  // Parameters: var_ind - global index of the unknown
  // Return value: local degrees of freedom on the coresponding 2D element for unknown "var_ind"
  DOF_iterator& Var_dof_2D(VAR var_ind) const {return container_2d->Var_dof(var_ind);}

  // Parameters: var_ind - global index of the unknown
  // Return value: local degrees of freedom on the coresponding 2D element
  //               for the right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof_2D(VAR var_ind) const {return container_2d->L_Var_dof(var_ind);}

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" for unknown "var_ind"
  DOF_iterator& Var_dof(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX
                 && temp().Var_dof[var_ind]!=0);
    temp().Var_dof[var_ind]->SetElement(ID());
    return *temp().Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  // Return value: local degrees of freedom on element "el_ind" 
  //               for right-hand side vector for unknown "var_ind"
  DOF_iterator& L_Var_dof(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX
                 && temp().L_Var_dof[var_ind]!=0);
    temp().L_Var_dof[var_ind]->SetElement(ID());
    return *temp().L_Var_dof[var_ind];
  }

  // Parameters: el_ind - element index on the face; var_ind - global index of the unknown
  //             j - coordinate index
  // Return value: local degrees of freedom on element "el_ind" 
  //               for flux vector for unknown "var_ind"
  DOF_iterator& Q_dof(VAR var_ind, int j) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX
                 && temp().Q_dof[var_ind][j]!=0);
    temp().Q_dof[var_ind][j]->SetElement(ID());
    return *temp().Q_dof[var_ind][j];
  }

#ifdef BAROCLINIC_KEY
#if TURBULENCE_KEY>1
  // Return value: local degrees of freedom on element "el_ind" for the z-component of the density gradient
  DOF_iterator& RHO_Z_dof() const
  {
    assert_debug(temp().Var_dof[RHO_Z]!=0);
    temp().Var_dof[RHO_Z]->SetElement(ID());
    return *temp().Var_dof[RHO_Z];
  }
//   DOF_iterator& RO_Z_dof() const
//   {
//     assert_debug(temp().Q_dof[RO][2]!=0);
//     temp().Q_dof[RO][2]->SetElement(ID());
//     return *temp().Q_dof[RO][2];
//   }
#endif
#endif

  // --------------- Accessors: Deal with data stored at quadrature points ---------------

  // Coriolis coefficient at the quadrature point, rvalue
  real Coriolis_quad() const {return container_2d->Coriolis_quad();}

  // Return value: depth at the quadrature point
  real depth_quad() const {return container_2d->depth_quad();}

  // Return value: surface elevation at the quadrature point
  real surface_elev_quad() const {return container_2d->surface_elev_quad();}

  // Parameters: i - coordinate index
  // Return value: tidal forcing term for the momentum equation
  real tidal_forcing(int i) const {return container_2d->TidalForcing(i);}

  // Parameters: var_ind - global index of the unknown
  // Return value: value of a 2D unknown at the quadrature point
  real Var_2D_quad(VAR var_ind) const {return container_2d->Var_quad(var_ind);}

#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  // Parameters: i - coordinate index
  // Return value: value of the depth integrated velocity at the quadrature point
  real UH_quad(int i) const {return container_2d->UH_quad(i);}
#endif

  // Parameters: var_ind - global index of the unknown
  // Return value: pointer to values of 2D basis functions at the quadrature point
  const real* basis_func_quad_2D(VAR var_ind) const
  {return container_2d->basis_func_quad(var_ind);}

  // Parameters: var_ind - global index of the unknown, j - coordinate index
  // Return value: pointer to values of gradients of 2D basis functions at the quadrature point
  const real* basis_grad_quad_2D(VAR var_ind, int j) const
  {return container_2d->basis_grad_quad(var_ind, j);}

  // Weight at the quadrature point, rvalue
  long_real weight_quad() const {return quad_local().weight;}

  // Weight at the quadrature point, lvalue
  long_real& weight_quad() {return quad_local().weight;}

  // Weight for 2D computations at the quadrature point, rvalue
  long_real elev_weight_quad() const
  {return weight_quad()*Mapping().GetVerticalScale()/(depth_quad()+surface_elev_quad());}

  // Weight for W computations at the quadrature point, rvalue
  long_real w_weight_quad() const
  {return weight_quad()/(depth_quad()+surface_elev_quad());}

  // Return value: coordinates of the quadrature point, rvalue
  const real *coord_quad() const {return quad_local().coord;}

  // Return value: coordinates of the quadrature point, lvalue
  real *coord_quad() {return quad_local().coord;}

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, rvalue
  real coord_quad(int i) const
  {
    assert_debug(i>=0 && i<d_space);
    return quad_local().coord[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th coordinate of the quadrature point, lvalue
  real& coord_quad(int i)
  {
    assert_debug(i>=0 && i<d_space);
    return quad_local().coord[i];
  }

  // Parameters: var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the quadrature point, rvalue
  const real* basis_func_quad(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && basis_index[var_ind]!=BAD_INDEX);
    return &mass_storage(quad_local().basis_func_ind[basis_index[var_ind]]);
  }

  // Parameters: var_ind - global index of the unknown
  // Return value: pointer to values of basis functions at the quadrature point, lvalue
  real* basis_func_assign_quad(int i)
  {
    assert_debug(i>=0 && i<n_basis);
    return &mass_storage(quad_local().basis_func_ind[i]);
  }

  // Parameters: var_ind - global index of the unknown, j - coordinate index
  // Return value: pointer to values of gradients of basis functions at the quadrature point, rvalue
  const real* basis_grad_quad(VAR var_ind, int j) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space &&
                 basis_index[var_ind]!=BAD_INDEX);
    return &mass_storage(quad_local().basis_grad_ind[basis_index[var_ind]][j]);
  }

  // Parameters: var_ind - global index of the unknown, j - coordinate index
  // Return value: pointer to values of gradients of basis functions at the quadrature point, lvalue
  real* basis_grad_assign_quad(int i, int j)
  {
    assert_debug(i>=0 && i<n_basis && j>=0 && j<d_space);
    return &mass_storage(quad_local().basis_grad_ind[i][j]);
  }

  // Parameters: var_ind - global index of the unknown
  // Return value: value of the unknown at the quadrature point, rvalue
  real Var_quad(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX);
    return Var_dof(var_ind)*basis_func_quad(var_ind);
  }

  // Return value: value of the "W" at the quadrature point, rvalue
  real W_quad() const {return Var_dof(W)*basis_func_quad(W);}

  // Parameters: var_ind - global index of the unknown; j - coordinate index
  // Return value: value of the flux of unknown at the quadrature point, rvalue
  real Q_quad(VAR var_ind, int j) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && j>=0 && j<d_space && 
                 VAR_3D_DT_INDEX_INV[var_ind]!=BAD_INDEX);
    return Q_dof(var_ind, j)*basis_func_quad(var_ind);
  }

//   // Parameters: var_ind - global index of the unknown
//   // Return value: the constant component only of the value of the unknown, rvalue
//   real Var_const(VAR var_ind) const
//   {
//     assert_debug(var_ind>=0 && var_ind<d_var && VAR_3D_INDEX_INV[var_ind]!=BAD_INDEX);
//     return Var_dof(var_ind)(0)*basis_func_quad(var_ind)[0];
//   }

#ifdef BAROCLINIC_KEY
//   // Return value: pointer to values of vertically integrated basis functions
//   //               at the quadrature point, rvalue
//   const real* basis_func_integrated_quad() const
//   {return &mass_storage(quad_local().basis_func_integrated_ind);}

//   // Return value: pointer to values of vertically integrated basis functions
//   //               at the quadrature point, lvalue
//   real* basis_func_integrated_quad()
//   {return &mass_storage(quad_local().basis_func_integrated_ind);}

//   // Return value: depth integrated density forcing term at the quadrature point
//   real density_forcing_integrated_quad() const {return container_2d->density_forcing_quad();}

  // Return value: pointer to values of vertically integrated basis functions
  //               at the quadrature point, rvalue
  const real* basis_func_integrated_quad() const
{
  static func_type basis_func_integrated;

  Basis_dof(RHO)->basis_func_vert_integral_phys(coord_quad(), Mapping(), basis_func_integrated);
  return basis_func_integrated;
}

  // Return value: density forcing at the quadrature point
  real density_forcing_quad() const
  {return Var_dof(RHO)*basis_func_integrated_quad()+Var_dof(RHO_FRS)*basis_func_quad_2D(XI);}

//   real density_forcing_quad() const {return Var_dof(RO)*basis_func_integrated_quad()+
//       Var_dof(RO_COMPLEMENT)*basis_func_quad_2D(RO_FORCING);}

// #if TURBULENCE_KEY>1
//   // Return value: z-component of the density gradient at the quadrature point
//   real density_grad_z_quad() const {return RHO_Z_dof()*basis_func_quad(RHO_Z);}
// #endif
#endif

  //----------------- Updaters: Update data stored at quadrature points ------------------

  // Sets the quadrature index for the corresponding 2d element
  // Parameters: ind - quadrature point index
  void SetIndex(int ind)
  {BaseStoragePrism::SetIndex(ind); container_2d->SetIndex(Quad()->GetSubIndex(0, ind));}

  // Increments the quadrature index for the corresponding 2d element
  // Parameters: i - dummy parameter
  void operator++(int i)
  {BaseStoragePrism::operator++(i); container_2d->SetIndex(Quad()->GetSubIndex(0, GetIndex()));}

  // Initialize storage for data stored at quadrature points
  virtual void InitializeStorage();

  // Update data shared by all prisms
  // Parameters: quad - pointer to quadrature object;
  //             basis - pointer to basis object
  //             var_new - global storage for degrees of freedom;
  //             var_flux_new - global storage for degrees of freedom for the fluxes;
  //             var_rhs_new - global storage for degrees of freedom for the right hand side
  static void Update_static(const QUADRATURE *quad, const BASIS *basis[d_var_3d], 
                            DOF var_new[d_var], DOF var_flux_new[d_var][d_space], 
                            DOF var_rhs_new[d_var]);

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  virtual void Update_basis_quad();

  // Numerics
  void Projection(real, void fn(real, const real[], real[]));
  void Projection_forcing(real, void fn(real, const real[], real[]));
  void Global_change_L2();
  void Global_error_L2(real, void fn(real, const real[], real[]));
  void ComputeDepthAveragedVelocity(real t);
  void ElementIntegration(real t);
  void ElementIntegrationDiffusion(real t);
  void ElementIntegrationDiffHoriz();
  void ElementIntegrationDiffVert();
  void ElementIntegrationVertVelocity();
  void ElementIntegrationVertVelocityMatrix();
#ifdef BAROCLINIC_KEY
//   void ElementIntegrationDensity();
  void ProjectionDensity();
#ifdef MASS_CONSERVATION_KEY
  void Tracer_Projection(real t);
  void Tracer_Mass();
#endif
#endif
#if TURBULENCE_KEY>1
  void LimitTurbulence();
#endif

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const PRISM_SW&);
};



//--------------------------------------------------------------------------------------
// PRISM_PARALLEL Class: Parallel prism class in 3D for shallow-water problems
//--------------------------------------------------------------------------------------

class PRISM_PARALLEL : public PRISM_SW
{
public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  PRISM_PARALLEL(TRIANGLE_2D_SW& elem_2d);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  PRISM_PARALLEL(std::ifstream& in);

  // Default destructor
  ~PRISM_PARALLEL() {}

  // Return value: pointer to a prism of the same type as the caller
  PRISM_PARALLEL* CreatePrism(TRIANGLE_2D_SW& face)
  {PRISM_PARALLEL* prism = new PRISM_PARALLEL(face); assert(prism!=0); return prism;}

  // Initialize storage for data stored at quadrature points
  void InitializeStorage();
};



//--------------------------------------------------------------------------------------
// PRISM_NONPARALLEL Class: Nonparallel prism class in 3D for shallow-water problems
//--------------------------------------------------------------------------------------

class PRISM_NONPARALLEL : public PRISM_SW
{
private:
  std::size_t mass_matrix_ind[d_var_3d];

public:
  // Default constructor
  // Parameters: elem_2d - corresponding 2D element
  PRISM_NONPARALLEL(TRIANGLE_2D_SW& elem_2d);

  // Constructor for hot start input
  // Parameters: in - hot start input file stream
  PRISM_NONPARALLEL(std::ifstream& in);

  // Default destructor
  ~PRISM_NONPARALLEL() {}

  // Compute mapping from the reference triangle
  void ComputeReferenceMapping();

  // Return value: pointer to a prism of the same type as the caller
  PRISM_NONPARALLEL* CreatePrism(TRIANGLE_2D_SW& face)
  {PRISM_NONPARALLEL* prism = new PRISM_NONPARALLEL(face); assert(prism!=0); return prism;}

  // Invert mass matrices for all basis functions
  void InvertMassMatrix();

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Refine the element
  // Return value: id of the first child element
  void Refine();

  // Unrefine the element
  void UnRefine();
#endif

  // Parameters: var_ind - global index of the unknown
  // Retrurn value: pointer to the mass matrix for "var_ind"
  const long_real* mass_matrix(VAR var_ind) const
  {
    assert_debug(var_ind>=0 && var_ind<d_var && basis_index[var_ind]!=BAD_INDEX);
    return (const long_real*)&mass_storage(mass_matrix_ind[basis_index[var_ind]]);
  }

  // Parameters: i - basis object index
  // Retrurn value: pointer to the mass matrix for i-th basis object
  long_real* mass_matrix_assign(int i)
  {
    assert_debug(i>=0 && i<n_basis);
    return (long_real*)&mass_storage(mass_matrix_ind[i]);
  }

  // Initialize storage for data stored at quadrature points
  void InitializeStorage();

  // Update rarely changed data at quadrature points (basis functions, coordinates, etc)
  void Update_basis_quad();

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream&, const PRISM_NONPARALLEL&);
};

#endif
