//----------------------------------------------------------------------------------------
// pde.cpp: Contains definitions of functions describing physical processes
//----------------------------------------------------------------------------------------

#include <cmath>
#include "general.H"
#include "pde.H"

extern PROBLEM_DESCRIPTION problem_info;

#if TURBULENCE_KEY==3  // two equation turbulence models
GLS_Data gls_data;     // Data for two equation turbulence models


// Model initialization
// Parameters: model - model choice parameter
void GLS_Data::Init(int model)
{
  switch(model)
    {
    case 5: // Mellor-Yamada k-kl model
      p=0.; m=1.; n=1.;
      sigma_k=sigma_psi=2.44;
      c1=0.9; c2=0.5; c3_unst=1.; c3_st=2.53;
      k_min=7.6e-6; psi_min=7.6e-7; // psi_min=1.e-8
      c_mu0=0.5544;
      break;
    case 6: // Burchard et al. k-eps model
      p=3.; m=1.5; n=-1.;
      sigma_k=1.; sigma_psi=1.3;
      c1=1.44; c2=1.92; c3_unst=1.; c3_st=-0.52;
      k_min=7.6e-6; psi_min=1.e-8; // psi_min=1.e-8
      c_mu0=0.5544;
      break;
    case 7: // Wilcox et al k-omega model
      p=-1.; m=0.5; n=-1.;
      sigma_k=2.; sigma_psi=2.;
      c1=0.555; c2=0.833; c3_unst=1.; c3_st=-0.58;
      k_min=7.6e-6; psi_min=5.e-3;
      c_mu0=0.5544;
      break;
    case 8: // Umlauf-Burchard, 2003, gen model
      p=2.; m=1.; n=-2./3.;
      sigma_k=0.8; sigma_psi=1.07;
      c1=1.; c2=1.22; c3_unst=1.; c3_st=0.1;
      k_min=7.6e-6; psi_min=1.e-5; // psi_min=5.e-6
      c_mu0=0.5544;
      break;
    default: std::cerr << "File pde.H: Undefined choice of the two-equation turbulence model!\n";
      std::exit(1);
    }
  C=std::pow(c_mu0, p);
  kappa=0.4;
  c_mu0_3=c_mu0*c_mu0*c_mu0;
  // Stability coefficients initialization
  A1=0.392; A2=17.07; A3=-6.127;
  B1=0.4939; B2=-30.19;
  Gh_min=-0.28; Gh_max=0.028; Gh_c=0.02;
}
#endif


//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//                                  PDE functions
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Advection term computation (for horizontal faces)
// Parameters: var - values of unknowns at the point; advec - result
void advection_horiz(const real var[d_var], long_real advec[d_var][d_space])
{
  advec[XI][0]=var[U];
  advec[XI][1]=var[V];
  advec[XI][2]=var[W];
  advec[U][0]=problem_info.advect_coef*var[U]*var[U]+problem_info.g*var[XI];
  advec[U][1]=problem_info.advect_coef*var[U]*var[V];
  advec[U][2]=problem_info.advect_coef*var[U]*var[W];
  advec[V][0]=problem_info.advect_coef*var[V]*var[U];
  advec[V][1]=problem_info.advect_coef*var[V]*var[V]+problem_info.g*var[XI];
  advec[V][2]=problem_info.advect_coef*var[V]*var[W];
}

// Advection term computation
// Parameters: var - values of unknowns at the point; advec - result
void advection(const real var[d_var], long_real advec[d_var][d_space])
{
  advec[XI][0]=var[XI]*var[U];
  advec[XI][1]=var[XI]*var[V];
  advec[XI][2]=0.;
  advec[U][0]=problem_info.advect_coef*var[U]*var[U]+problem_info.g*var[XI];
  advec[U][1]=problem_info.advect_coef*var[U]*var[V];
  advec[U][2]=problem_info.advect_coef*var[U]*var[W];
  advec[V][0]=problem_info.advect_coef*var[V]*var[U];
  advec[V][1]=problem_info.advect_coef*var[V]*var[V]+problem_info.g*var[XI];
  advec[V][2]=problem_info.advect_coef*var[V]*var[W];
}

// Transport term computation
// Parameters: var - values of unknowns at the point; advec - result
void transport(const real var[d_var], long_real advec[d_var][d_space])
{
  static const VAR* ind=VAR_TRANSPORT_INDEX;

  for (int i=0; i<d_var_transport; i++)
  {
    advec[ind[i]][0]=var[ind[i]]*var[U];
    advec[ind[i]][1]=var[ind[i]]*var[V];
    advec[ind[i]][2]=var[ind[i]]*var[W];
  }
}

#if !defined(TEST_KEY) && defined(BAROCLINIC_KEY)
// Equation of state for density: B. Klinger:http://mason.gmu.edu/~bklinger/seawater.pdf
// Parameters: coord - point coordinates, var - values of unknowns at the point
// Return value: density divided by the reference density
long_real density(const real coord[d_space], const real var[d_var])
{
  long_real p = problem_info.Rho_dep_pres ? 0.001*(var[XI]-coord[2])/problem_info.unit_conversion_factor : 0.,
    alpha=0.0708*(1.+0.351*p+0.068*(1.-0.0683*p)*var[T]), beta=0.808-0.0085*p, 
    gamma=0.003*(1.-0.059*p-0.012*(1.-0.064*p)*var[T]), Rho_0=problem_info.unit_conversion_factor*
    problem_info.unit_conversion_factor*problem_info.unit_conversion_factor*problem_info.Rho_0;

  return (999.83+(5.053-0.048*p)*p+beta*var[S]-alpha*var[T]-gamma*(35.-var[S])*var[T])-Rho_0;
}
#endif

// Vertical eddy viscosity/diffusivity coefficients computation (constant coefficients)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_const(const real param[], real diff[d_var][d_space])
{
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=problem_info.diff_vert[ind];
  }
}

// Vertical eddy viscosity/diffusivity coefficients computation (Davies ~ U*H)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_uh(const real param[], real diff[d_var][d_space])
{
  real H=param[0], U=param[1], V=param[2], coef=std::sqrt(U*U+V*V)*H;
  //       coef = H>25. ? 25. : H;
  //       coef*=problem_info.vert_eddy_visc*std::sqrt(U*U+U*U);

  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=coef*problem_info.diff_vert[ind];
  }
}

// Vertical eddy viscosity/diffusivity coefficients computation (Davies ~ U*U)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_uu(const real param[], real diff[d_var][d_space])
{
  real U=param[1], V=param[2], coef=U*U+V*V;

  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=coef*problem_info.diff_vert[ind];
  }
}

#if TURBULENCE_KEY>1
// Right hand side computation for Xing-Davies k turbulence model
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_k(const real param[], real diff[d_var][d_space], real *rhs)
{
  static const VAR* ind=VAR_TRANSPORT_INDEX;
  static const real kappa=0.4, c_mu0=std::sqrt(0.3), c_mu0_3=c_mu0*c_mu0*c_mu0;
  real k_min=1.e-6*problem_info.unit_conversion_factor*problem_info.unit_conversion_factor;
  real dz_top=kappa*param[0], dz_bot=kappa*param[1], N_2=param[2];
  real k = param[5]>k_min ? param[5] : k_min, sqrt_k=std::sqrt(k);;
  real u_z=param[3], v_z=param[4], u_term=u_z*u_z+v_z*v_z;
  real Ri = u_term<1.e-4 ? 0. : N_2/u_term;
  real F_l = Ri<0 ? std::pow(1.-14.*Ri, 0.25) : std::exp(-2.3*Ri);
  real l=F_l*dz_bot*std::sqrt(dz_top/(dz_bot+dz_top));
  real v_t=c_mu0*sqrt_k*l;
  v_t = v_t>problem_info.diff_vert_max ? problem_info.diff_vert_max : v_t;
  real v_t_prime=v_t/0.7, v_k=v_t;

  diff[U][2]=v_t*problem_info.diff_vert[U];
  diff[V][2]=v_t*problem_info.diff_vert[V];
  for (int i=0; i<d_var_transport; i++)
    diff[ind[i]][2]=v_t_prime*problem_info.diff_vert[ind[i]];
  diff[K][2]=v_k*problem_info.diff_vert[K];
  if (rhs==0)
    return;

  // Compute the right-hand-side for turbulence variables
  real eps=c_mu0_3*k*sqrt_k/l;
  rhs[K]=v_t*u_term+v_t_prime*N_2-eps;
}
#endif


#if TURBULENCE_KEY==3
// Right hand side computation for the GLS model (Warner et al., 2005)
// Parameters: param - parameters; diff - result
void diffusion_vert_coef_gls(const real param[], real diff[d_var][d_space], real *rhs)
{
  static const VAR* ind=VAR_TRANSPORT_INDEX;
  static const real sqrt_2=std::sqrt(2.);
  real S_M, S_H, F_wall, v_t, v_t_prime, v_k, v_psi;
  real dz_top=gls_data.kappa*param[0], dz_bot=gls_data.kappa*param[1], N_2=param[2];
  real k = param[5]>gls_data.k_min ? param[5] : gls_data.k_min, sqrt_k=std::sqrt(k);
  real psi = param[6]>gls_data.psi_min ? param[6] : gls_data.psi_min;
  real l_max = (N_2>1.e-5 && 0.56*k<4.*dz_bot*dz_bot*N_2) ? std::sqrt(0.56*k/N_2) : 2.*dz_bot;
  real l=std::pow(k, real(-gls_data.m/gls_data.n))*std::pow(psi/gls_data.C, real(1./gls_data.n));
  if (l>l_max)
    {
      l=l_max;
      psi=gls_data.C*std::pow(k, real(gls_data.m))*std::pow(l, real(gls_data.n));
    }

  gls_data.Stab(N_2*l*l/2./k, S_M, S_H);
  v_t=sqrt_2*S_M*sqrt_k*l;
  v_t = v_t>problem_info.diff_vert_max ? problem_info.diff_vert_max : v_t;
  v_t_prime=S_H/S_M*v_t, v_k=v_t/gls_data.sigma_k, v_psi=v_t/gls_data.sigma_psi;

  diff[U][2]=v_t*problem_info.diff_vert[U];
  diff[V][2]=v_t*problem_info.diff_vert[V];
  for (int i=0; i<d_var_transport; i++)
    diff[ind[i]][2]=v_t_prime*problem_info.diff_vert[ind[i]];
  diff[K][2]=v_k*problem_info.diff_vert[K];
  diff[E][2]=v_psi*problem_info.diff_vert[E];
  if (rhs==0)
    return;

  // Compute the right-hand-side for turbulence variables
  real u_z=param[3], v_z=param[4], u_term=u_z*u_z+v_z*v_z;
  real c3 = N_2<0. ? gls_data.c3_unst : gls_data.c3_st;
  real eps=gls_data.c_mu0_3*k*sqrt_k/l;
  if (gls_data.n==1) // Wall proximity function for the Mellor-Yamada model
    F_wall=1.+l*l*(1.33/dz_bot/dz_bot);//+0.25/dz_top/dz_top);
  else
    F_wall=1.;
  rhs[K]=v_t*u_term+v_t_prime*N_2-eps;
  rhs[E]=psi/k*(gls_data.c1*v_t*u_term+c3*v_t_prime*N_2-gls_data.c2*eps*F_wall);
}
#endif

// Horizontal eddy viscosity/diffusivity coefficients computation (constant coefficients)
// Parameters: param - parameters; diff - result
void diffusion_horiz_coef(const real param[], real diff[d_var][d_space])
{
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=problem_info.diff_horiz[ind];
    diff[ind][1]=problem_info.diff_horiz[ind];
  }
}


//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//                                   Riemann solvers
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Normal advective flux solver for Riemann problem on vertical faces
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void advection_vert_flux(const long_real normal[d_space], const real var_l[d_var],
                         const real var_r[d_var], long_real flux[d_var])
{
  long_real advec[d_var][d_space];

  advection(var_r, advec);
  for (int i=0; i<d_space_2d; i++)
    {
      flux[XI]+=advec[XI][i]*normal[i];
      flux[U]+=advec[U][i]*normal[i];
      flux[V]+=advec[V][i]*normal[i];
    }
}

// HLL solver for Riemann problem on vertical faces
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void HLL(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
         long_real flux[d_var])
{
  long_real u_l_n=var_l[U]*normal[0]+var_l[V]*normal[1];
  long_real u_r_n=var_r[U]*normal[0]+var_r[V]*normal[1];
  long_real u_star=0.5*(u_l_n+u_r_n), gh_l=std::sqrt(0.25*u_l_n*u_l_n+problem_info.g*var_l[XI]);
  long_real gh_r=std::sqrt(0.25*u_r_n*u_r_n+problem_info.g*var_r[XI]), gh_star=0.5*(gh_l+gh_r); 
  long_real s_l, s_r, advec_l[d_var][d_space], advec_r[d_var][d_space];

  flux[XI]=flux[U]=flux[V]=0.;
  s_l = 1.5*u_l_n-gh_l>1.5*u_star-gh_star ? 1.5*u_star-gh_star : 1.5*u_l_n-gh_l;
  if (s_l>=0.)
    {
      advection(var_l, advec_l);
      for (int i=0; i<d_space_2d; i++)
        {
          flux[XI]+=advec_l[XI][i]*normal[i];
          flux[U]+=advec_l[U][i]*normal[i];
          flux[V]+=advec_l[V][i]*normal[i];
        }
    }
  else
    {
      advection(var_r, advec_r);
      s_r = 1.5*u_r_n+gh_r<1.5*u_star+gh_star ? 1.5*u_star+gh_star : 1.5*u_r_n+gh_r;
      if (s_r<=0.)
        for (int i=0; i<d_space_2d; i++)
          {
            flux[XI]+=advec_r[XI][i]*normal[i];
            flux[U]+=advec_r[U][i]*normal[i];
            flux[V]+=advec_r[V][i]*normal[i];
          }
      else
        {
          advection(var_l, advec_l);
          for (int i=0; i<d_space_2d; i++)
            {
              flux[XI]+=(s_r*advec_l[XI][i]-s_l*advec_r[XI][i])*normal[i];
              flux[U]+=(s_r*advec_l[U][i]-s_l*advec_r[U][i])*normal[i];
              flux[V]+=(s_r*advec_l[V][i]-s_l*advec_r[V][i])*normal[i];
            }
          flux[XI]+=s_l*s_r*(var_r[XI]-var_l[XI]);
          flux[U]+=s_l*s_r*(var_r[U]-var_l[U]);
          flux[V]+=s_l*s_r*(var_r[V]-var_l[V]);
          flux[XI]/=s_r-s_l;
          flux[U]/=s_r-s_l;
          flux[V]/=s_r-s_l;
        }
    }
}

//-------------------------------------------------------------------------

// HLLC solver for Riemann problem on vertical faces
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void HLLC(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
          long_real flux[d_var])
{
  long_real u_l_n=var_l[U]*normal[0]+var_l[V]*normal[1];
  long_real u_r_n=var_r[U]*normal[0]+var_r[V]*normal[1];
  long_real u_star=0.5*(u_l_n+u_r_n), gh_l=std::sqrt(problem_info.g*var_l[XI]);
  long_real gh_r=std::sqrt(problem_info.g*var_r[XI]), gh_star=0.5*(gh_l+gh_r);
  long_real s_l, s_r, s_m, advec_l[d_var][d_space], advec_r[d_var][d_space];

  flux[XI]=flux[U]=flux[V]=0.;
  s_l = u_l_n-gh_l>u_star-gh_star ? u_star-gh_star : u_l_n-gh_l;
  if (s_l>=0.)
    {
      advection(var_l, advec_l);
      for (int i=0; i<d_space_2d; i++)
        {
          flux[XI]+=advec_l[XI][i]*normal[i];
          flux[U]+=advec_l[U][i]*normal[i];
          flux[V]+=advec_l[V][i]*normal[i];
        }
    }
  else
    {
      s_r = u_r_n+gh_r<u_star+gh_star ? u_star+gh_star : u_r_n+gh_r;
      if (s_r<=0.)
        {
          advection(var_r, advec_r);
          for (int i=0; i<d_space_2d; i++)
            {
              flux[XI]+=advec_r[XI][i]*normal[i];
              flux[U]+=advec_r[U][i]*normal[i];
              flux[V]+=advec_r[V][i]*normal[i];
            }
        }
      else
        {
//   s_m=u_star;
          s_m=(u_r_n*(s_r-u_r_n)-u_l_n*(s_l-u_l_n)-problem_info.g*(var_r[XI]-var_l[XI]))/
            (s_r-u_r_n-s_l+u_l_n);
          if (s_m>=0.)
            {
              long_real coef_star=(s_m-u_l_n)/(s_l-s_m);
              flux[XI]=s_l*coef_star*var_l[XI];
              flux[U]=s_l*coef_star*(var_l[U]+problem_info.g*normal[0]*var_l[XI]/(s_l-s_m));
              flux[V]=s_l*coef_star*(var_l[V]+problem_info.g*normal[1]*var_l[XI]/(s_l-s_m));
              advection(var_l, advec_l);
              for (int i=0; i<d_space_2d; i++)
                {
                  flux[XI]+=advec_l[XI][i]*normal[i];
                  flux[U]+=advec_l[U][i]*normal[i];
                  flux[V]+=advec_l[V][i]*normal[i];
                }
            }
          else
            {
              long_real coef_star=(s_m-u_r_n)/(s_r-s_m);
              flux[XI]=s_r*coef_star*var_r[XI];
              flux[U]=s_r*coef_star*(var_r[U]+problem_info.g*normal[0]*var_r[XI]/(s_r-s_m));
              flux[V]=s_r*coef_star*(var_r[V]+problem_info.g*normal[1]*var_r[XI]/(s_r-s_m));
              advection(var_r, advec_r);
              for (int i=0; i<d_space_2d; i++)
                {
                  flux[XI]+=advec_r[XI][i]*normal[i];
                  flux[U]+=advec_r[U][i]*normal[i];
                  flux[V]+=advec_r[V][i]*normal[i];
                }
            }
        }
    }
}

//-------------------------------------------------------------------------

// Roe's solver for Riemann problem on vertical faces
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void Roe(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
         long_real flux[d_var])
{
  long_real u_hat[d_space_2d]={0.5*(var_l[U]+var_r[U]), 0.5*(var_l[V]+var_r[V])};
  long_real H_hat=0.5*(var_l[XI]+var_r[XI]), u_n=u_hat[0]*normal[0]+u_hat[1]*normal[1];
  long_real a_hat=std::sqrt(u_n*u_n+4.*problem_info.g*H_hat), advec[d_var][d_space];

  flux[XI]=flux[U]=flux[V]=0.;
  advection(var_l, advec);
  for (int i=0; i<d_space_2d; i++)
    {
      flux[XI]+=advec[XI][i]*normal[i];
      flux[U]+=advec[U][i]*normal[i];
      flux[V]+=advec[V][i]*normal[i];
    }
  if (3.*u_n<a_hat)
    {
      long_real u_jump[d_space_2d]={var_r[U]-var_l[U], var_r[V]-var_l[V]};
      long_real u_jump_n=u_jump[0]*normal[0]+u_jump[1]*normal[1];
      long_real H_quot=(var_r[XI]-var_l[XI])/H_hat, plus_hat, minus_hat, alfa;

      plus_hat=0.5*(a_hat+u_n);
      alfa=0.5*(3.*u_n-a_hat)*(plus_hat*H_quot-u_jump_n)/a_hat;
      flux[U]+=alfa*(u_hat[0]-plus_hat*normal[0]);
      flux[V]+=alfa*(u_hat[1]-plus_hat*normal[1]);
      flux[XI]+=alfa*H_hat;
      if (u_n<0.)
        {
          alfa=u_n*(u_jump[1]*normal[0]-u_jump[0]*normal[1]-
                    H_quot*(u_hat[1]*normal[0]-u_hat[0]*normal[1]));
          flux[U]-=alfa*normal[1];
          flux[V]+=alfa*normal[0];
          if (3.*u_n<-a_hat)
            {
              minus_hat=0.5*(a_hat-u_n);
              alfa=0.5*(3.*u_n+a_hat)*(minus_hat*H_quot+u_jump_n)/a_hat;
              flux[U]+=alfa*(u_hat[0]+minus_hat*normal[0]);
              flux[V]+=alfa*(u_hat[1]+minus_hat*normal[1]);
              flux[XI]+=alfa*H_hat;
            }
        }
    }
}

//-------------------------------------------------------------------------

// Roe's solver for Riemann problem on vertical faces with an entropy fix
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void RoeEntropyFix(const long_real normal[d_space], const real var_l[d_var],
                   const real var_r[d_var], long_real flux[d_var])
{
  long_real u_hat[d_space_2d]={0.5*(var_r[U]+var_l[U]), 0.5*(var_r[V]+var_l[V])};
  long_real u_jump[d_space_2d]={var_r[U]-var_l[U], var_r[V]-var_l[V]};
  long_real u_n=u_hat[0]*normal[0]+u_hat[1]*normal[1];
  long_real u_jump_n=u_jump[0]*normal[0]+u_jump[1]*normal[1];
  long_real H_hat=0.5*(var_l[XI]+var_r[XI]), H_jump=var_r[XI]-var_l[XI];
  long_real a_hat=std::sqrt(u_n*u_n+4.*problem_info.g*H_hat), 
    plus_hat=a_hat+u_n, minus_hat=a_hat-u_n, advec[d_var][d_space];

  flux[XI]=flux[U]=flux[V]=0.;
  advection(var_l, advec);
  for (int i=0; i<d_space_2d; i++)
    {
      flux[XI]+=advec[XI][i]*normal[i];
      flux[U]+=advec[U][i]*normal[i];
      flux[V]+=advec[V][i]*normal[i];
    }

  long_real u_cur[3]={var_l[U], var_l[V], var_l[XI]}, u_n_tmp=u_cur[0]*normal[0]+u_cur[1]*normal[1];
  long_real a_hat_tmp=std::sqrt(u_n*u_n+4.*problem_info.g*u_cur[2]), eig_val[3];
  long_real eig_vect[3][3]={
    {(u_hat[0]-0.5*plus_hat*normal[0])/H_hat, (u_hat[1]-0.5*plus_hat*normal[1])/H_hat, 1.}, 
    {-normal[1], normal[0], 0.}, 
    {(u_hat[0]+0.5*minus_hat*normal[0])/H_hat, (u_hat[1]+0.5*minus_hat*normal[1])/H_hat, 1.}};
  long_real alfa[3]={(0.5*plus_hat*H_jump-H_hat*u_jump_n)/a_hat, u_jump[1]*normal[0]-
                u_jump[0]*normal[1]-H_jump*(u_hat[1]*normal[0]-u_hat[0]*normal[1])/H_hat, 
                (0.5*minus_hat*H_jump+H_hat*u_jump_n)/a_hat}, eig_val_l, eig_val_r;

  // Eigenvalue 1
  eig_val_l=0.5*(3.*u_n_tmp-a_hat_tmp);
  for (int j=0; j<3; u_cur[j]+=alfa[0]*eig_vect[0][j], j++);
  u_n_tmp=u_cur[0]*normal[0]+u_cur[1]*normal[1]; 
  a_hat_tmp=std::sqrt(u_n*u_n+4.*problem_info.g*u_cur[2]);
  if (eig_val_l<0.)
    {
      eig_val_r=0.5*(3.*u_n_tmp-a_hat_tmp);
      eig_val[0] = eig_val_r>0. ? 
        eig_val_l*(eig_val_r-0.5*(3.*u_n-a_hat))/(eig_val_r-eig_val_l) : 0.5*(3.*u_n-a_hat);
    }
  else
    eig_val[0]=0.;
  // Eigenvalue 2
  eig_val_l=u_n_tmp;
  for (int j=0; j<3; u_cur[j]+=alfa[1]*eig_vect[1][j], j++);
  u_n_tmp=u_cur[0]*normal[0]+u_cur[1]*normal[1]; 
  a_hat_tmp=std::sqrt(u_n*u_n+4.*problem_info.g*u_cur[2]);
  if (eig_val_l<0.)
    {
      eig_val_r=u_n_tmp;
      eig_val[1] = eig_val_r>0. ? eig_val_l*(eig_val_r-u_n)/(eig_val_r-eig_val_l) : u_n;
    }
  else
    eig_val[1]=0.;
  // Eigenvalue 3
  eig_val_l=0.5*(3.*u_n_tmp+a_hat_tmp);
  for (int j=0; j<3; u_cur[j]+=alfa[2]*eig_vect[2][j], j++);
  u_n_tmp=u_cur[0]*normal[0]+u_cur[1]*normal[1]; 
  a_hat_tmp=std::sqrt(u_n*u_n+4.*problem_info.g*u_cur[2]);
  if (eig_val_l<0.)
    {
      eig_val_r=0.5*(3.*u_n_tmp+a_hat_tmp);
      eig_val[2] = eig_val_r>0. ? 
        eig_val_l*(eig_val_r-0.5*(3.*u_n+a_hat))/(eig_val_r-eig_val_l) : 0.5*(3.*u_n+a_hat);
    }
  else
    eig_val[2]=0.;
  
  for (int j=0; j<3; j++)
    {
      flux[XI]+=alfa[j]*eig_val[j]*eig_vect[j][2];
      flux[U]+=alfa[j]*eig_val[j]*eig_vect[j][0];
      flux[V]+=alfa[j]*eig_val[j]*eig_vect[j][1];
    }
}


//-------------------------------------------------------------------------

// Lax-Friedrichs solver for Riemann problem on vertical faces
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void LaxFriedrichs(const long_real normal[d_space], const real var_l[d_var],
                   const real var_r[d_var], long_real flux[d_var])
{
  long_real u_hat[d_space_2d]={0.5*(var_r[U]+var_l[U]), 0.5*(var_r[V]+var_l[V])};
  long_real  H_hat=0.5*(var_r[XI]+var_l[XI]), u_n=u_hat[0]*normal[0]+u_hat[1]*normal[1], 
    coef, advec_l[d_var][d_space], advec_r[d_var][d_space];

  coef=1.5*std::abs(u_n)+0.5*std::sqrt(u_n*u_n+4.*problem_info.g*H_hat);
  flux[XI]=-coef*(var_r[XI]-var_l[XI]);
  flux[U]=-coef*(var_r[U]-var_l[U]);
  flux[V]=-coef*(var_r[V]-var_l[V]);
  advection(var_l, advec_l);
  advection(var_r, advec_r);
  for (int i=0; i<d_space_2d; i++)
    {
      flux[XI]+=0.5*(advec_l[XI][i]+advec_r[XI][i])*normal[i];
      flux[U]+=0.5*(advec_l[U][i]+advec_r[U][i])*normal[i];
      flux[V]+=0.5*(advec_l[V][i]+advec_r[V][i])*normal[i];
    }
}


//-------------------------------------------------------------------------

// FORCE solver for Riemann problem on vertical faces
// Parameters: normal - normal to the face;
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
void FORCE(const long_real normal[d_space], const real var_l[d_var], const real var_r[d_var],
           long_real flux[d_var])
{
  real u_m[d_var];
  long_real u_hat[d_space_2d]={0.5*(var_r[U]+var_l[U]), 0.5*(var_r[V]+var_l[V])};
  long_real H_hat=0.5*(var_r[XI]+var_l[XI]);
  long_real u_n=u_hat[0]*normal[0]+u_hat[1]*normal[1], flux_l[d_var], flux_r[d_var],
    coef, advec_m[d_var][d_space], advec_l[d_var][d_space], advec_r[d_var][d_space];

  flux_l[XI]=flux_l[U]=flux_l[V]=flux_r[XI]=flux_r[U]=flux_r[V]=0.;
  coef=1.5*std::abs(u_n)+0.5*std::sqrt(u_n*u_n+4.*problem_info.g*H_hat);
  advection(var_l, advec_l);
  advection(var_r, advec_r);
  for (int i=0; i<d_space_2d; i++)
    {
      flux_l[XI]+=advec_l[XI][i]*normal[i];
      flux_l[U]+=advec_l[U][i]*normal[i];
      flux_l[V]+=advec_l[V][i]*normal[i];
      flux_r[XI]+=advec_r[XI][i]*normal[i];
      flux_r[U]+=advec_r[U][i]*normal[i];
      flux_r[V]+=advec_r[V][i]*normal[i];
    }
  u_m[XI]=H_hat-0.5/coef*(flux_r[XI]-flux_l[XI]);
  u_m[U]=u_hat[0]-0.5/coef*(flux_r[U]-flux_l[U]);
  u_m[V]=u_hat[1]-0.5/coef*(flux_r[V]-flux_l[V]);
  advection(u_m, advec_m);
  flux[XI]=0.25*(flux_l[XI]+flux_r[XI]-coef*(var_r[XI]-var_l[XI]));
  flux[U]=0.25*(flux_l[U]+flux_r[U]-coef*(var_r[U]-var_l[U]));
  flux[V]=0.25*(flux_l[V]+flux_r[V]-coef*(var_r[V]-var_l[V]));
  for (int i=0; i<d_space_2d; i++)
    {
      flux[XI]+=0.5*advec_m[XI][i]*normal[i];
      flux[U]+=0.5*advec_m[U][i]*normal[i];
      flux[V]+=0.5*advec_m[V][i]*normal[i];
    }
}

