//----------------------------------------------------------------------------------------
// runge_kutta.cpp: Definitions for main time stepping functions
//----------------------------------------------------------------------------------------

//#include <iomanip>
#include <iostream>
#include <cmath>
#include "dof.H"
#include "runge_kutta.H"
#include "les_solver.H"
#include "pde.H"
#include "mesh_container.H"

#ifdef TEST_KEY
#include "test.H"
#endif


// Global variables
extern DOF var_stage[max_stages][d_var];  // Global storage for degrees of freedom for Runge-Kutta stages
extern DOF var_rhs[d_var];                // Global storage for degrees of freedom for the right hand side
extern DOF var_flux[d_var][d_space];      // Global storage for degrees of freedom for fluxes
extern DOF var[d_var];                    // Global storage for degrees of freedom for unknowns
extern PROBLEM_DESCRIPTION problem_info;  // Storage for physical parameters
extern MESH mesh;                         // Mesh object

// Runge-Kutta method coefficients for various schemes
const long_real RUNGE_KUTTA::a_table[d_rk][max_stages][max_stages]=
{{{1.}},
{{1., 0.}, {0.5, 0.5}},
{{1., 0., 0.}, {3./4., 1./4., 0.}, {1./3., 0., 2./3.}}};
const long_real RUNGE_KUTTA::b_table[d_rk][max_stages]=
{{1.}, {1., 0.5}, {1., 1./4., 2./3.}};
const long_real RUNGE_KUTTA::d_table[d_rk][max_stages]=
{{0.}, {0., 1.}, {0., 1., 1./2.}};

//------------------------------------------------
// Time stepping function
// Parameters: t_0 - time; dt - time step
//------------------------------------------------
void RUNGE_KUTTA::time_stepping(long_real t_0, long_real dt)
//------------------------------------------------
{
  int n_stages=problem_info.runge_kutta_method;

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
  // Update mesh geometry: adaptivity
  if ((problem_info.mesh_update!=0) && (mesh.UpdateCounter()<=1) && 
      (problem_info.max_refinement>0) && (t_0>problem_info.refinement_start))
    mesh.UpdateMesh();
#endif

  // Time stepping
  for (int i=0; i<n_stages; i++)
  {
    real t=t_0+d[i]*dt;

    // Compute tidal forcing
    mesh.ComputeTidalForcing(t);

    // Update mesh geometry: free surface
    if ((problem_info.mesh_update!=0) && (--mesh.UpdateCounter()<=0))
    {
      mesh.UpdateSurfaceElevation(t);
      mesh.UpdateCounter()=problem_info.mesh_update*n_stages;

      // Update quantities stored at quadrature points
      for (std::size_t j=0; j<mesh.prism_list.size(); j++)
        if (mesh.prism_list[j].IsActive() && !mesh.prism_list[j].IsUpdated())
          mesh.prism_list[j].Update_basis_quad();
      for (std::size_t j=0; j<mesh.face_horiz_list.size(); j++)
        if (mesh.face_horiz_list[j].IsActive() && !mesh.face_horiz_list[j].IsUpdated())
          mesh.face_horiz_list[j].Update_basis_quad();
      for (std::size_t j=0; j<mesh.face_vert_list.size(); j++)
        if (mesh.face_vert_list[j].IsActive() && !mesh.face_vert_list[j].IsUpdated())
          mesh.face_vert_list[j].Update_basis_quad();
    }

    // Zero storage for degrees of freedom
    for (int j=0; j<d_var_dt; j++)
    {
      var_rhs[VAR_DT_INDEX[j]]=0.;
      for (int k=0; k<d_space; k++)
        var_flux[VAR_DT_INDEX[j]][k]=0.;
    }
    for (int j=0; j<d_var_c; j++)
      var[VAR_C_INDEX[j]]=0.;

#ifdef MASS_CONSERVATION_KEY
  PRISM_SW::mass_change()=0.;
#endif

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY) // Initialize adaptivity data
    for (std::size_t j=0; j<mesh.triangle_2d_list.size(); j++)
    {
      TRIANGLE_2D_SW& element_2d=mesh.triangle_2d_list[j];
      element_2d.UnSetAdaptivitySwitches();
      if (!element_2d.IsActive())
      {
        if (element_2d.IsRefined())
          element_2d.SetToBeUnRefined();
        if (element_2d.Parent()!=BAD_INDEX)
          mesh.triangle_2d_list[element_2d.Parent()].UnSetToBeUnRefined();
      }
    }
#endif
#ifdef ADAPTIVITY_H_2D_KEY
    for (std::size_t j=0; j<mesh.triangle_2d_list.size(); j++)
      if (mesh.triangle_2d_list[j].IsActive())
        mesh.triangle_2d_list[j].InitializeAdaptivityData();
#endif
#ifdef ADAPTIVITY_H_3D_KEY
    for (std::size_t j=0; j<mesh.prism_list.size(); j++)
      if (mesh.prism_list[j].IsActive())
        mesh.prism_list[j].InitializeAdaptivityData();
#endif

    for (std::size_t j=0; j<mesh.triangle_2d_list.size(); j++)
      if (mesh.triangle_2d_list[j].IsActive())
      {
        // Check the minimum depth condition
        mesh.triangle_2d_list[j].MinDepthCheck(problem_info.min_depth);
        // Update values of state variables at quadrature points
        mesh.triangle_2d_list[j].Update_variables_quad(t);
      }
    // Update values of state variables at quadrature points on edges
    for (std::size_t j=0; j<mesh.edge_2d_list.size(); j++)
      if (mesh.edge_2d_list[j].IsActive())
        mesh.edge_2d_list[j].Update_variables_quad(t);

#if TURBULENCE_KEY>1  // one and two equation turbulence models
    // Limit minimum values of turbulence variables
    LimitTurbulenceVariables(t);
#endif

    // .......................
    // Compute advection terms
    Func_advection(t);

    // .......................
    // Compute flux variables
    Func_diffusion(t);

    // Perform necessary outputs
    if (i==0)
      mesh.Output(t_0, 2);

    // Multiplication with the mass matrix where necessary
    for (std::size_t j=0; j<mesh.prism_nonparallel_list.size(); j++)
    {
      PRISM_NONPARALLEL& prism=mesh.prism_nonparallel_list[j];
      if (prism.IsActive())
        for (int l=0; l<d_var_3d_dt; l++)
          LU_substitution(prism.L_Var_dof(VAR_3D_DT_INDEX[l]).GetDim(),
                          prism.mass_matrix(VAR_3D_DT_INDEX[l]), 
                          &prism.L_Var_dof(VAR_3D_DT_INDEX[l]));
    }

    // Compute the next Runge-Kutta stage
    for (int j=0; j<d_var_dt; j++)
    {
      VAR ind=VAR_DT_INDEX[j];

      if (i<n_stages-1)
        var_stage[i][ind]=var[ind];
      var[ind]*=a[i][i];
      var[ind].Add(dt*b[i], var_rhs[ind]);
      for (int k=0; k<i; k++)
        var[ind].Add(a[i][k], var_stage[k][ind]);
    }
#ifdef MASS_CONSERVATION_KEY
    PRISM_SW::mass_stage(i)=PRISM_SW::mass();
    PRISM_SW::mass()*=a[i][i];
    PRISM_SW::mass()+=dt*b[i]*PRISM_SW::mass_change();
    for (int k=1; k<i; k++)
      PRISM_SW::mass()+=a[i][k]*PRISM_SW::mass_stage(k);
#endif
  }
  mesh.IterationCounter()++;
  if (problem_info.hot_output!=0 && (mesh.IterationCounter()%problem_info.hot_freq==0))
    mesh.HotOutput(t_0+dt);
}

//-----------------------------------------------------------
// The right-hand side function for the time-stepping routine
// Parameters: t_0 - time
//-----------------------------------------------------------
void Func_advection(real t)
//-----------------------------------------------------------
{
#ifdef TEST_KEY // Projection of the forcing function for analytical test problems
  IterateOverList(t, mesh.triangle_2d_list, solution_analytic_elevation_forcing, 
                  &TRIANGLE_2D_SW::Projection_elevation_forcing);
  IterateOverList(t, mesh.prism_list, solution_analytic_forcing, &PRISM_SW::Projection_forcing);
#endif

#ifdef BAROCLINIC_KEY // Computation of the baroclinic forcing term
  ComputeDensityForcing(t);
#endif

#ifdef MASS_CONSERVATION_KEY
  IterateOverList(t, mesh.prism_list, &PRISM_SW::Tracer_Projection);
#endif

#ifdef DEPTH_AVERAGED_VELOCITY_KEY // Compute depth averaged horizontal velocity components
  IterateOverList(t, mesh.prism_list, &PRISM_SW::ComputeDepthAveragedVelocity);
#endif

  // Integration over vertical faces
  IterateOverList(t, mesh.face_vert_list, &FACE_VERT_SW::FaceVertIntegration);

  // Vertical velocity computation
  ComputeVerticalVelocity(t);

  // Integration over elements
  IterateOverList(t, mesh.prism_list, &PRISM_SW::ElementIntegration);
}

//------------------------------------------------
// Computation of the auxilliary variables in the mixed formulation
// Parameters: t_0 - time
//------------------------------------------------
void Func_diffusion(real t)
//------------------------------------------------
{
  if (problem_info.diff_vert_switch!=0)
  {
    // Multiplication with mass matrix where necessary
    for (std::size_t i=0; i<mesh.prism_nonparallel_list.size(); i++)
    {
      PRISM_NONPARALLEL& prism=mesh.prism_nonparallel_list[i];
      if (prism.IsActive())
        {
          for (int l=0; l<problem_info.n_diff_vert; l++)
            LU_substitution(prism.Q_dof(problem_info.diff_vert_ind[l], 2).GetDim(),
                            prism.mass_matrix(problem_info.diff_vert_ind[l]),
                            &prism.Q_dof(problem_info.diff_vert_ind[l], 2));
#ifdef BAROCLINIC_KEY
#if TURBULENCE_KEY>1  // one and two equation turbulence models
          LU_substitution(prism.RHO_Z_dof().GetDim(), prism.mass_matrix(RHO), &prism.RHO_Z_dof());
#endif
#endif
        }
    }
  }
  if (problem_info.diff_horiz_switch!=0)
  {
    // Multiplication with mass matrix where necessary
    for (std::size_t i=0; i<mesh.prism_nonparallel_list.size(); i++)
    {
      PRISM_NONPARALLEL& prism=mesh.prism_nonparallel_list[i];
      if (prism.IsActive())
        for (int l=0; l<problem_info.n_diff_horiz; l++)
          for (int k=0; k<d_space_2d; k++)
            LU_substitution(prism.Q_dof(problem_info.diff_horiz_ind[l], k).GetDim(),
                            prism.mass_matrix(problem_info.diff_horiz_ind[l]),
                            &prism.Q_dof(problem_info.diff_horiz_ind[l], k));
    }
  }
  // Add diffusion term contribution if necessary
  if (problem_info.diff_vert_switch!=0 || problem_info.diff_horiz_switch!=0)
  {
    // Integration over elements - horizontal direction
    IterateOverList(t, mesh.prism_list, &PRISM_SW::ElementIntegrationDiffusion);
    // Integration over horizontal faces
    IterateOverList(t, mesh.face_horiz_list, &FACE_HORIZ_SW::FaceHorizIntegrationDiffusion);
    // Integration over vertical faces
    IterateOverList(t, mesh.face_vert_list, &FACE_VERT_SW::FaceVertIntegrationDiffusion);
  }
}

//------------------------------------------------
// Computation of the vertical velocity component
// Parameters: t - time
//------------------------------------------------
void ComputeVerticalVelocity(real t)
//------------------------------------------------
{
  // Integration over elements for W 
  // horizontal faces for momentum and turbulence
  //
  for (std::size_t i=0; i<mesh.face_horiz_list.size(); i++)
    if (mesh.face_horiz_list[i].IsActive() && mesh.face_horiz_list[i].IsBottom())
    {
      PRISM_SW *prism=&mesh.prism_list[mesh.face_horiz_list[i].Element(0)];

      while(prism!=0)
      {
        int dim=prism->Var_dof(W).GetDim();
        FACE_HORIZ_SW &top_face=mesh.face_horiz_list[prism->TopFace()], 
          &bot_face=mesh.face_horiz_list[prism->BottomFace()];

        // Update W matrix if needed for U terms
        if (prism->w_matrix()[0]==0.)
        {
          IterateOverQuadrature(*prism, &PRISM_SW::ElementIntegrationVertVelocityMatrix);
          IterateOverQuadrature(top_face, &FACE_HORIZ_SW::FaceHorizIntegrationVertVelocityTopMatrix);
          if (LU_decomposition(dim, prism->w_matrix())==BAD_INDEX)
            std::cerr << "Singular W matrix!" << std::endl;
        }

        IterateOverQuadrature(t, bot_face, &FACE_HORIZ_SW::FaceHorizIntegration);
        IterateOverQuadrature(*prism, &PRISM_SW::ElementIntegrationVertVelocity);
        if (top_face.IsSurface())
          IterateOverQuadrature(top_face, &FACE_HORIZ_SW::FaceHorizIntegrationVertVelocityTop);

        // Solve for W-vector
        LU_substitution(dim, prism->w_matrix(), &prism->Var_dof(W));

        if (mesh.face_horiz_list[prism->TopFace()].IsSurface())
          IterateOverQuadrature(t, top_face, &FACE_HORIZ_SW::FaceHorizIntegration);

        // Go up to the next prism
        if (top_face.IsElement(0))
          prism=&mesh.prism_list[top_face.Element(0)];
        else
          prism=0;
      }
    }
}

#ifdef BAROCLINIC_KEY
//------------------------------------------------
// Computation of the baroclinic forcing term
// Parameters: t - time
//------------------------------------------------
void ComputeDensityForcing(real t)
//------------------------------------------------
{
  // Integrating density from the free surface to the current point
  for (std::size_t i=0; i<mesh.face_horiz_list.size(); i++)
    if (mesh.face_horiz_list[i].IsActive() && mesh.face_horiz_list[i].IsSurface())
    {
      PRISM_SW *prism=&mesh.prism_list[mesh.face_horiz_list[i].Element(1)];

      while(prism!=0)
      {
        FACE_HORIZ_SW &top_face=mesh.face_horiz_list[prism->TopFace()], 
          &bot_face=mesh.face_horiz_list[prism->BottomFace()];

        IterateOverQuadrature(*prism, &PRISM_SW::ProjectionDensity);

        // Multiplication with mass matrix where necessary
        if (!prism->IsParallel())
          LU_substitution(prism->Var_dof(RHO).GetDim(), ((PRISM_NONPARALLEL*)prism)->mass_matrix(RHO), 
                          &prism->Var_dof(RHO));

        IterateOverQuadrature(top_face, &FACE_HORIZ_SW::ProjectionDensityFaceHoriz);
        if (bot_face.IsBottom())
          IterateOverQuadrature(bot_face, &FACE_HORIZ_SW::ProjectionDensityFaceHoriz);

        // Go down to the next prism
        if (bot_face.IsElement(1))
          prism=&mesh.prism_list[bot_face.Element(1)];
        else
          prism=0;
      }
    }
}
#endif

#if TURBULENCE_KEY>1  // one and two equation turbulence models
//------------------------------------------------
// Limiting of minimum values of turbulence variables
// Parameters: t - time
//------------------------------------------------
void LimitTurbulenceVariables(real t)
//------------------------------------------------
{
  for (std::size_t i=0; i<mesh.prism_list.size(); i++)
    if (mesh.prism_list[i].IsActive())
    {
      PRISM_SW &prism=mesh.prism_list[i];

      IterateOverQuadrature(prism, &PRISM_SW::LimitTurbulence);

      // Multiplication with mass matrix where necessary
      if (!prism.IsParallel())
      {
        LU_substitution(prism.L_Var_dof(K).GetDim(), ((PRISM_NONPARALLEL&)prism).mass_matrix(K), 
                        &prism.L_Var_dof(K));
#if TURBULENCE_KEY==3  // two equation turbulence models
        LU_substitution(prism.L_Var_dof(E).GetDim(), ((PRISM_NONPARALLEL&)prism).mass_matrix(E), 
                        &prism.L_Var_dof(E));
#endif
      }
    }
  var[K].Add(1.,var_rhs[K]);
  var_rhs[K]=0.;
#if TURBULENCE_KEY==3  // two equation turbulence models
  var[E].Add(1.,var_rhs[E]);
  var_rhs[E]=0.;
#endif
}
#endif
