//----------------------------------------------------------------------------------------
// storage_container.H: Storage for data at quadrature points
//----------------------------------------------------------------------------------------

#ifndef STORAGE_CONTAINER_H
#define STORAGE_CONTAINER_H

#include <cassert>
#include <vector>
#include "general.H"


//----------------------------------------------------------------------------------------
// StorageContainer template class: Storage for data at quadrature points
//----------------------------------------------------------------------------------------


// Template parameters: A - data shared by all objects of this type
//                      B - data at quadrature points shared by all objects of this type
//                      D - local data at quadrature points
template <typename A, typename B, typename D>
class StorageContainer
{
private:
  unsigned char quad_index, n_quad;  // local quadrature index and number of quadrature points
  std::size_t global_index;                  // global index for the object

protected:
  static A temp_storage;                     // data shared by all objects of this type
  static B quad_temp_storage[max_quad];      // data at quadrature points shared by all objects of this type
  static std::vector<D> quad_local_storage;  // local data at quadrature points
  static std::vector<real> mass_storage_;    // storage for variable length data

  // Return value: storage for A
  A& temp() const {return temp_storage;}

  // Return value: storage for B
  B& quad_temp() const 
  {
    assert_debug(quad_index<n_quad && IsInitialized());
    return quad_temp_storage[quad_index];
  }

  // Return value: storage for D
  D& quad_local() const 
  {
    assert_debug(quad_index<n_quad && IsInitialized());
    return quad_local_storage[global_index+quad_index];
  }

  // Parameters: i - mass storage index
  // Return value: data stored at i in mass_storage
  real& mass_storage(std::size_t i) const
  {
    assert_debug(i<mass_storage_.size());
    return mass_storage_[i];
  }

public:
  // Default constructor
  StorageContainer() {global_index=(std::size_t)BAD_INDEX; quad_index=255; n_quad=0;}

  // Default destructor
  ~StorageContainer() {}

  // Initialize storage
  // Parameters: n - number of quadrature points
  void Initialize(int n)
  {
    D empty_value;
    assert(n>0 && n<256);
    if (sizeof(D)==0) return;
    n_quad=(unsigned char)n;
    global_index=quad_local_storage.size();
    quad_local_storage.insert(quad_local_storage.end(), n, empty_value);
  }

  // Return value: is storage initialized?
  bool IsInitialized() const {return global_index!=(std::size_t)BAD_INDEX;}

  // Return value: is storage updated?
  bool IsUpdated() const {return quad_index!=255;}

  // Unset update flag
  void UnSetUpdated() {quad_index=255;}

  // Set local quadrature index
  // Parameters: ind - new index
  void SetIndex(int ind)
  {
    assert_debug(IsInitialized() && ind>=0 && ind<=n_quad);
    quad_index=(unsigned char)ind;
  }

  // Return value: local quadrature index
  unsigned char GetIndex() const {return quad_index;}

  // Return value: is local quadrature index in range?
  bool InRange() const {return quad_index<n_quad;}

  // Increment local quadrature index
  void operator++(int i) {quad_index++;}
};

// Empty structure definitions (can be used in place of A, B, or D parameter)
typedef struct {} empty;

// Static members definitions
template <typename A, typename B, typename D>
A StorageContainer<A, B, D>::temp_storage;
template <typename A, typename B, typename D>
B StorageContainer<A, B, D>::quad_temp_storage[max_quad];
template <typename A, typename B, typename D>
std::vector<D> StorageContainer<A, B, D>::quad_local_storage;
template <typename A, typename B, typename D>
std::vector<real> StorageContainer<A, B, D>::mass_storage_;


//--------------------------------------------------------------------------------------
//               Functions for looping over quadrature points and objects
//--------------------------------------------------------------------------------------

// Loop over quadrature points
template <typename T>
inline void IterateOverQuadrature(T& obj, void (T::*func)()) 
{
  for (obj.SetIndex(0); obj.InRange(); obj++)
    (obj.*func)();
}

// Loop over quadrature points
template <typename T>
inline void IterateOverQuadrature(real t, T& obj, void f(real, const real[], real[]), 
                                  void (T::*func)(real, void fn(real, const real[], real[]))) 
{
  for (obj.SetIndex(0); obj.InRange(); obj++)
    (obj.*func)(t, f);
}

// Loop over quadrature points
template <typename T>
inline void IterateOverQuadrature(real t, T& obj, void (T::*func)(real)) 
{
  for (obj.SetIndex(0); obj.InRange(); obj++)
    (obj.*func)(t);
}

// Loop over objects
template <typename T, typename L>
inline void IterateOverList(L& list, void (T::*func)()) 
{
  for (std::size_t i=0; i<list.size(); i++)
    if (list[i].IsActive())
      IterateOverQuadrature<T>(list[i], func);
}

// Loop over objects
template <typename T, typename L>
inline void IterateOverList(real t, L& list, void f(real, const real[], real[]), 
                            void (T::*func)(real, void fn(real, const real[], real[])))
{
  for (std::size_t i=0; i<list.size(); i++)
    if (list[i].IsActive())
      IterateOverQuadrature<T>(t, list[i], f, func);
}

// Loop over objects
template <typename T, typename L>
inline void IterateOverList(real t, L& list, void (T::*func)(real))
{
  for (int i=0; i<list.size(); i++)
    if (list[i].IsActive())
      IterateOverQuadrature<T>(t, list[i], func);
}

// // Loop over objects
// template <typename T, typename L>
// inline void IterateOverList(L& list, void (T::*func)()) 
// {
//   L::iterator iter;
//   for (iter=list.begin(); iter!=list.end(); iter++)
//     if (iter->IsActive())
//       IterateOverQuadrature<T>(*iter, func);
// }

#endif
